import abce


class Killer(abce.Agent, abce.Household):
    def init(self):
        # your agent initialization goes here, not in __init__
        pass

    def kill_silent(self):
        return self.round

    def kill_loud(self):
        return self.round



from flexx.pyscript import window
from flexx.ui import Layout
from flexx import event


class DockPanel(Layout):
    """ A layout that displays its children as dockable widgets.

    This is a high level layout allowing the user to layout the child
    widgets as he/she likes. The title of each child is used for its
    corresponding tab label.

    the widget's style property 'location' determines where to place the
    next new widget.

    location = N, S, W, or E:
        for absolute placement.
    location = L(eft), R(ight), U(nder), O(ver) B(efore) A(fter):
        for relative placement.
    relative:
        Relative to the last widget or to the widget set with set_relative
    """

    CSS = """

    div.p-TabBar {
        border: 1px solid #ccc;
        width: 20% !important;
        height: 100% !important;
        overflow-y: scroll;
        border-color: blue !important;
    }

    div.p-TabBar-body {
        border-color: blue !important;
    }

    ul.p-TabBar-content {
    display: block  !important;
        background-color: blue !important;
    }

    span.p-TabBar-tabText {
        color: white !important;
    }

    span.p-TabBar-tabText:hover {
        color: black !important;
    }

    li.p-TabBar-tab {
        border-color: blue !important;
        background-color: blue !important;
    }

    li.p-TabBar-tab:hover {
        border-color: black !important;
        border: 2px solid #ccc;
        background-color: DodgerBlue !important;
    }

    li.p-mod-current {
        background-color: DodgerBlue  !important;
    }
    div.p-TabBar-tab button {
        display: block  !important;
        background-color: red !important;
        color: black;
        padding: 22px 16px;
        width: 20% !important;
        border: none;
        outline: none;
        text-align: left !important;
        cursor: pointer !important;
        transition: 0.3s;
    }

    div.p-StackedPanel {
        top: 0px !important;
        padding: 0px 12px;
        border: 1px solid #ccc;
        width: 80% !important;
        border-left: none !important;
        left: 20% !important;
        height: 100% !important;
    }
    """

    def selectWidget(self, widget):
        self.emit('myselectWidget', {'widget': widget})

    class Both:

        def set_relative(self, widget):
            """ The next attached widget will be placed relative to
            this one """
            self.relative = widget.phosphor

    class JS:
        @event.connect('myselectWidget')
        def myselectWidget(self, event):
            self.phosphor.selectWidget(event['widget'].phosphor)

        def _init_phosphor_and_node(self):
            self.phosphor = window.phosphor.dockpanel.DockPanel()
            self.node = self.phosphor.node
            self.relative = None

        def _add_child(self, widget):
            after = widget.style.split('location:')[1]
            try:
                location = after.split(';')[0]
            except TypeError:
                location = ""
            location = location.upper()
            if 'W' in location:
                self.phosphor.insertLeft(widget.phosphor)
            elif 'N' in location:
                self.phosphor.insertTop(widget.phosphor)
            elif 'E' in location:
                self.phosphor.insertRight(widget.phosphor)
            elif 'S' in location:
                self.phosphor.insertBottom(widget.phosphor)
            else:
                if self.relative is None:
                    self.phosphor.insertLeft(widget.phosphor)
                elif 'L' in location:
                    try:
                        self.phosphor.insertLeft(widget.phosphor,
                                                 self.relative)
                        self.lastworking = self.relative
                    except Exception:
                        self.phosphor.insertLeft(widget.phosphor,
                                                 self.lastworking)

                elif 'O' in location:
                    try:
                        self.phosphor.insertTop(widget.phosphor,
                                                self.relative)
                        self.lastworking = self.relative
                    except Exception:
                        self.phosphor.insertTop(widget.phosphor,
                                                self.lastworking)

                elif 'R' in location:
                    try:
                        self.phosphor.insertRight(widget.phosphor,
                                                  self.relative)
                        self.lastworking = self.relative
                    except Exception:
                        self.phosphor.insertRight(widget.phosphor,
                                                  self.lastworking)

                elif 'U' in location:
                    try:
                        self.phosphor.insertBottom(widget.phosphor,
                                                   self.relative)
                        self.lastworking = self.relative
                    except Exception:
                        self.phosphor.insertBottom(widget.phosphor,
                                                   self.lastworking)

                elif 'B' in location:
                    try:
                        self.phosphor.insertTabBefore(widget.phosphor,
                                                      self.relative)
                        self.lastworking = self.relative
                    except Exception:
                        self.phosphor.insertTabBefore(widget.phosphor,
                                                      self.lastworking)

                elif 'A' in location:
                    try:
                        self.phosphor.insertTabAfter(widget.phosphor,
                                                     self.relative)
                        self.lastworking = self.relative
                    except Exception:
                        self.phosphor.insertTabAfter(widget.phosphor,
                                                     self.lastworking)

                else:
                    self.phosphor.insertLeft(widget.phosphor)
            self.relative = widget.phosphor



"""Form to input parameters acording to the parameter_mask"""
import abce
from flexx import ui, event


def form(parameter_mask, names):
    """Gererates Form class instance with parameter_mask and names"""
    class Form(ui.Widget):
        def init(self):
            self.fields = {}
            self.radio_buttons = {}
            self.result_property = {}
            self.slider_to_textfield = {}
            self.textfield_to_slider = {}
            self.int_sliders = set()
            self.sliders = []

            with ui.GroupWidget(title="Simulation parameter"):
                ui.Label(text="scroll down to start",
                         style="float: right; color: blue", wrap=True)
                for parameter, value in list(parameter_mask.items()):
                    try:
                        title = names[parameter]
                    except KeyError:
                        title = parameter

                    if isinstance(value, bool):
                        self.fields[parameter] = ui.CheckBox(text=title)
                        self.result_property[parameter] = 'checked'

                    elif isinstance(value, list):
                        self.radio_buttons[parameter] = {}
                        with ui.GroupWidget(title=title,
                                            style="width:fit-content;"):
                            for option in value:
                                self.radio_buttons[parameter][option] = \
                                    ui.RadioButton(text=option)
                        self.radio_buttons[parameter][value[0]].checked = True

                    else:
                        if isinstance(value, tuple):
                            min_value, default, max_value = sorted(value)
                        elif isinstance(value, (int, float)):
                            min_value, default, max_value = 0, value, value * 2

                        is_integer = False
                        if isinstance(value, (int, float, tuple)):
                            if (isinstance(default, int) and
                                    isinstance(max_value, int)):
                                step = 1
                                is_integer = True
                                # if default is float, type is float
                                if isinstance(default, float):
                                    step = (max_value - min_value) / 100
                            else:
                                step = (max_value - min_value) / 100
                            with ui.Widget():
                                ui.Label(text=title, wrap=True)
                                slider = ui.Slider(min=min_value,
                                                   max=max_value,
                                                   value=default,
                                                   step=step)
                                lineeditor = ui.LineEdit(title=title,
                                                         text=default)
                                self.sliders.append((slider, lineeditor))
                            self.fields[parameter] = slider
                            self.result_property[parameter] = 'value'
                            slider.connect('value', self.stt)
                            lineeditor.connect('submit', self.tts)
                            self.slider_to_textfield[slider] = lineeditor
                            self.textfield_to_slider[lineeditor] = slider
                            if is_integer:
                                self.int_sliders.add(slider)

                        elif isinstance(value, str):
                            with ui.Widget():
                                ui.Label(text=title,
                                         wrap=True, style="width: 80%")
                                self.fields[parameter] = \
                                    ui.LineEdit(title=title,
                                                text=value,
                                                style='width: 95%;')
                            self.result_property[parameter] = 'text'
                        elif value is None:
                            ui.Label(text=title, wrap=True)
                        else:  # field
                            print(str(value) + "not recognized")
                with ui.VBox():
                    self.btn = ui.Button(text="start simulation")
                    self.repeat_execution_checker = ui.ToggleButton(
                        text='repeated execution')
                with ui.GroupWidget(title="Save"):
                    with ui.HBox():
                        self.name = ui.LineEdit(title="Name:",
                                                placeholder_text='name')
                        self.save = ui.Button(text="Save Parameters")
                    self.description = ui.LineEdit(
                        title="Description",
                        text='',
                        style='width: 95%;',
                        placeholder_text='description')

        def parse_parameter(self):
            parameter = {}
            for key, element in self.fields.items():
                parameter[key] = getattr(element,
                                         self.result_property[key])
                if element in self.int_sliders:
                    parameter[key] = int(parameter[key])
            for parameter, group in self.radio_buttons.items():
                for value, checkbox in group.items():
                    if checkbox.checked:
                        parameter[parameter] = value
            return parameter

        @event.connect('save.mouse_click')
        def _save(self, events):
            parameter = self.parse_parameter()
            parameter['name'] = self.name.text
            parameter['description'] = self.description.text
            abce.parameter_database.upsert(parameter, keys=['name'])
            self.emit('update_parameter_database', parameter)

        def load_parameter(self, event):
            parameter = abce.parameter_database.find_one(name=event['name'])

            for key, element in self.fields.items():
                if isinstance(element, ui.CheckBox):
                    element.checked = parameter[key]
                elif isinstance(element, ui.LineEdit):
                    element.text = parameter[key]
                elif isinstance(element, ui.Slider):
                    element.value = parameter[key]
                    self.slider_to_textfield[element].text = parameter[key]

            for parameter, group in self.radio_buttons.items():
                for value, checkbox in group.items():
                    checkbox.checked = bool(parameter[parameter] == value)
            return parameter

        @event.connect('btn.mouse_click')
        def wdg(self, *event):
            parameter = self.parse_parameter()
            self.emit('run_simulation',
                      {'simulation_parameter': parameter})

        @event.connect('repeat_execution_checker.mouse_click')
        def repeat_execution(self, events):
            parameter = self.parse_parameter()
            self.emit('_repeat_execution', {'simulation_parameter': parameter})

        def stt(self, events):  # This is executed in python, but should
                        # be executed in JS without server interaction
            slider = events['source']
            if slider in self.int_sliders:
                self.slider_to_textfield[slider].text = int(
                    events['new_value'])
            else:
                self.slider_to_textfield[slider].text = float(
                    events['new_value'])

        def tts(self, events):  # This is executed in python, but should
                        # be executed in JS without serve interaction
            slider = self.textfield_to_slider[events['source']]
            if slider in self.int_sliders:
                new_value = int(events['source'].text)
            else:
                new_value = float(events['source'].text)
            if new_value > slider.max:
                slider.max = new_value
            if new_value < slider.min:
                slider.min = new_value
            slider.value = new_value

        class JS:
            @event.connect('_repeat_execution')
            def _repeat_execution(self, event):
                if self.repeat_execution_checker.checked:
                    self.emit('repeatexecution', event)
    return Form


def assert_all_of_the_same_type(values):
    for item in values:
        if isinstance(item, type(values[0])):
            raise ValueError("all list values must be of the same type. "
                             "If 5.5 is used 0.0 instead of 0 must be used: " +
                             str(values))



""" Copyright 2012 Davoud Taghawi-Nejad

 Module Author: Davoud Taghawi-Nejad

 ABCE is open-source software. If you are using ABCE for your research you are
 requested the quote the use of this software.

 Licensed under the Apache License, Version 2.0 (the "License"); you may not
 use this file except in compliance with the License and quotation of the
 author. You may obtain a copy of the License at
       http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 License for the specific language governing permissions and limitations under
 the License.
"""
# pylint: disable=W0212, C0111


import multiprocessing as mp
from multiprocessing.managers import BaseManager
import traceback
from collections import defaultdict, ChainMap


class MyManager(BaseManager):
    pass


class ProcessorGroup:
    def __init__(self, batch, queues, processes):
        self.agents = {}
        self.batch = batch
        self.queues = queues
        self.queue = queues[self.batch]
        self.processes = processes

    def new_group(self, group):
        """ Creates a new group. """
        self.agents[group] = []

    def insert_or_append(self, group, ids, free_ids, Agent, simulation_parameters, agent_parameters, agent_arguments):
        """appends an agent to a group """
        if isinstance(agent_parameters, int):
            agent_parameters = ([] for _ in range(agent_parameters))

        for ap in agent_parameters:
            if free_ids:
                id = free_ids.popleft()
                ids[id] = id
            else:
                id = len(ids)
                ids.append(id)
            if id % self.processes == self.batch:
                agent = Agent(id, simulation_parameters, ap, **agent_arguments)
                agent._send = agent._send_multiprocessing
                agent._out = defaultdict(list)
                agent.init(**ChainMap(simulation_parameters, ap))
                agent._processes = self.processes
                if len(self.agents[group]) <= id // self.processes:
                    self.agents[group].append(agent)
                else:
                    self.agents[group][id // self.processes] = agent
        return ids

    def advance_round(self, time):
        for agents in self.agents.values():
            for agent in agents:
                if agent is not None:
                    agent._advance_round(time)

    def delete_agents(self, group, ids):
        for id in ids:
            if id % self.processes == self.batch:
                assert self.agents[group][id // self.processes].id == id
                assert self.agents[group][id // self.processes].group == group
                self.agents[group][id // self.processes] = None

    def do(self, groups, ids, command, args, kwargs):
        try:
            self.rets = []
            self.post = [[] for _ in range(self.processes)]
            for group, iss in zip(groups, ids):
                for i in iss:
                    if i is not None:
                        if i % self.processes == self.batch:
                            agent = self.agents[group][i // self.processes]
                            ret = agent._execute(command, args, kwargs)
                            self.rets.append(ret)
                            pst = agent._post_messages_multiprocessing(self.processes)
                            for o in range(self.processes):
                                self.post[o].extend(pst[o])
        except Exception:
            traceback.print_exc()
            raise

    def post_messages(self, groups, ids):
        for i in range(self.processes):
            self.queues[i].put(self.post[i])

        for i in range(self.processes):
            for receiver_group, receiver_id, envelope in self.queue.get():
                self.agents[receiver_group][receiver_id // self.processes].inbox.append(envelope)
        return self.rets

    def keys(self):
        return self.agents.keys()


class MultiProcess(object):
    """ This is a container for all agents. It exists only to allow for multiprocessing with MultiProcess.
    """
    def __init__(self, processes):
        manager = mp.Manager()
        self.queues = [manager.Queue() for _ in range(processes)]
        self.pool = mp.Pool(processes)
        MyManager.register('ProcessorGroup', ProcessorGroup)
        self.processor_groups = []
        self.managers = []
        for i in range(processes):
            manager = MyManager()
            manager.start()
            self.managers.append(manager)
            pg = manager.ProcessorGroup(i, self.queues, processes)
            self.processor_groups.append(pg)

    def new_group(self, group):
        """ Creates a new group. """
        for pg in self.processor_groups:
            pg.new_group(group)

    def insert_or_append(self, group, ids, free_ids, Agent, simulation_parameters, agent_parameters, agent_arguments):
        """appends an agent to a group """
        ids = self.pool.map(insert_or_append_wrapper, jkk(self.processor_groups,
                                                          group,
                                                          ids,
                                                          free_ids,
                                                          Agent,
                                                          simulation_parameters,
                                                          agent_parameters,
                                                          agent_arguments))
        return ids[0]

    def delete_agents(self, group, ids):
        self.pool.map(delete_agents_wrapper, jkk(self.processor_groups, group, ids))

    def do(self, groups, ids, command, args, kwargs):
        self.pool.map(wrapper, jkk(self.processor_groups, groups, ids, command, args, kwargs))

    def post_messages(self, groups, ids):
        ret = self.pool.map(post_messages, jkk(self.processor_groups, groups, ids))
        return sort_ret(ret)

    def advance_round(self, time):
        self.pool.map(advance_round_wrapper, jkk(self.processor_groups, time))

    def group_names(self):
        return self.processor_groups[0].keys()


def wrapper(args):
    pg, groups, ids, command, args, kwargs = args
    pg.do(groups, ids, command, args, kwargs)


def post_messages(args):
    pg, groups, ids = args
    return pg.post_messages(groups, ids)


def insert_or_append_wrapper(arg):
    pg, group, ids, free_ids, Agent, simulation_parameters, agent_parameters, agent_arguments = arg
    return pg.insert_or_append(group, ids, free_ids, Agent, simulation_parameters, agent_parameters, agent_arguments)


def delete_agents_wrapper(arg):
    pg, group, ids = arg
    pg.delete_agents(group, ids)


def advance_round_wrapper(arg):
    pg, time = arg
    pg.advance_round(time)


def sort_ret(ret):
    sorted_ret = []
    for i in range(len(ret[0])):
        for pg in ret:
            try:
                sorted_ret.append(pg[i])
            except IndexError:
                pass
    return sorted_ret


def jkk(iterator, *args):
    for i in iterator:
        yield (i,) + args



import abce


class ContractBuyerStop(abce.Agent, abce.Contracting):
    def init(self, rounds):
        self.last_round = rounds - 1
        if self.id == 0:
            self.create('labor_endowment', 1)

    def request_offer(self):
        if self.id == 1:
            if self.round % 10 == 0:
                self.given_contract = self.request_good_contract('contractbuyerstop', 0,
                                                                 'labor',
                                                                 quantity=5,
                                                                 price=10,
                                                                 duration=31)
            if self.round % 5 == 0 and self.round % 10 != 0:
                self.end_contract(self.given_contract)

    def accept_offer(self):
        if self.id == 0:
            contracts = self.get_contract_offers('labor')
            for contract in contracts:
                self.accepted_contract = self.accept_contract(contract)

    def deliver(self):
        for contract in self.contracts_to_deliver('labor'):
            self.deliver_contract(contract)
            assert self['labor'] == 0

    def pay(self):
        for contract in self.contracts_to_receive('labor'):
            if self.was_delivered_this_round(contract):
                self.create('money', 50)
                self.pay_contract(contract)
                assert self['money'] == 0

    def control(self):
        if self.round % 10 < 5:
            if self.id == 1:
                assert self.was_delivered_this_round(
                    self.given_contract), self.given_contract
                assert self['money'] == 0
                assert self['labor'] == 5
            else:
                assert self.was_paid_this_round(
                    self.accepted_contract), self._contracts_payed
                assert self['labor'] == 0, self['labor']
                assert self['money'] == 50, self['money']
                self.destroy('money')
        else:
            if self.id == 1:
                assert self['money'] == 0
                assert self['labor'] == 0
            else:
                assert self['labor'] == 5, self['labor']
                assert self['money'] == 0, self['money']
                self.destroy('labor')

    def clean_up(self):
        pass

    def all_tests_completed(self):
        if self.round == self.last_round and self.id == 0:
            print('Test request_offer      \t\t\tOK')
            print('Test end_contract       \t\t\tOK')



from .inventory import Inventory


class Goods:
    """ Each agent can access his goods. self['good_name'] shows the quantity of goods of a certain type an agent
    owns. Goods can be a string or any other python object.
    """
    def __init__(self, id, agent_parameters, simulation_parameters, group, trade_logging,
                 database, check_unchecked_msgs, expiring, perishable, resource_endowment, start_round=None):
        self._inventory = Inventory((group, id))
        self._resources = []

    def possession(self, good):
        """ returns how much of good an agent possesses.

        Returns:
            A number.

        possession does not return a dictionary for self.log(...), you can use
        self.possessions([...]) (plural) with self.log.

        Example::

            if self['money'] < 1:
                self.financial_crisis = True

            if not(is_positive(self['money']):
                self.bancrupcy = True

        """
        print("depreciated use self[good] or self.not_reserved[good]")
        return self._inventory[good]

    def possessions(self):
        """ returns all possessions """
        return self._inventory.possessions()

    def create(self, good, quantity):
        """ creates quantity of the good out of nothing

        Use create with care, as long as you use it only for labor and
        natural resources your model is macro-economically complete.

        Args:
            'good': is the name of the good
            quantity: number
        """
        self._inventory.create(good, quantity)

    def create_timestructured(self, good, quantity):
        """ creates quantity of the time structured good out of nothing.
        For example::

            self.creat_timestructured('capital', [10,20,30])

        Creates capital. 10 units are 2 years old 20 units are 1 year old
        and 30 units are new.

        It can alse be used with a quantity instead of an array. In this
        case the amount is equally split on the years.::

            self.create_timestructured('capital', 60)

        In this case 20 units are 2 years old 20 units are 1 year old
        and 20 units are new.

        Args:
            'good':
                is the name of the good

            quantity:
                an arry or number
        """
        self._inventory.create_timestructured(good, quantity)

    def _declare_expiring(self, good, duration):
        """ creates a good that has a limited duration
        """
        self._inventory._declare_expiring(good, duration)

    def not_reserved(self, good):
        """ Returns the amount of goods that are not reserved for a trade

        Args:
            good
        """
        return self._inventory.not_reserved(good)

    def destroy(self, good, quantity=None):
        """ destroys quantity of the good. If quantity is omitted destroys all

        Args::

            'good':
                is the name of the good
            quantity (optional):
                number

        Raises::

            NotEnoughGoods: when goods are insufficient
        """
        self._inventory.destroy(good, quantity)

    def _register_resource(self, resource, units, product):
        self._resources.append((resource, units, product))

    def _register_perish(self, good):
        self._inventory._perishable.append(good)

    def __getitem__(self, good):
        return self._inventory.haves[good]



import abce


class MyAgent(abce.Agent):
    def init(self, sp, ap):
        print('init', ap)
        self.mine = ap
        self.diff = False

    def call_me_maybe(self):
        print(self.name, self.mine)


class DifferentAgent(abce.Agent):
    def init(self, sp, ap):
        print('init', ap)
        self.mine = ap
        self.diff = True

    def call_me_maybe(self):
        print('I am different', self.name, self.mine)


sim = abce.Simulation()

myagent = sim.build_agents(MyAgent, 'myagent',
                           agent_parameters=[{'agent_parameters': 0},
                                             {'agent_parameters': 1},
                                             {'agent_parameters': 2},
                                             {'agent_parameters': 3},
                                             {'agent_parameters': 4},
                                             {'agent_parameters': 5},
                                             {'agent_parameters': 6}])

differentagent = sim.build_agents(DifferentAgent, 'differentagent',
                                  agent_parameters=[{'agent_parameters': 0},
                                                    {'agent_parameters': 1},
                                                    {'agent_parameters': 2},
                                                    {'agent_parameters': 3},
                                                    {'agent_parameters': 4},
                                                    {'agent_parameters': 5},
                                                    {'agent_parameters': 6}])

print('len', len(myagent))

sim.advance_round(0)
myagent.call_me_maybe()
print('simple')

for r in range(1, 14):
    sim.advance_round(r)
    myagent[r % 7].call_me_maybe()
    print('--')

print('two individuals')
for r in range(14, 28):
    (myagent[r % 7] + myagent[3]).call_me_maybe()
    print('--')

print('one class')
for r in range(28, 42):
    sim.advance_round(r)
    myagent[r % 7, 3].call_me_maybe()
    print('--')

print('whole class plus individual')
for r in range(42, 56):
    sim.advance_round(r)
    (myagent + myagent[r % 7]).call_me_maybe()
    print('--')

print('two classes')
for r in range(56, 70):
    sim.advance_round(r)
    (myagent + differentagent).call_me_maybe()
    print('--')

sim.finalize()



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# ABCE is open-source software. If you are using ABCE for your research you
# are requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
The :class:`abceagent.Agent` class is the basic class for creating your agent.
It automatically handles the
possession of goods of an agent. In order to produce/transforme goods you
need to also subclass the :class:`abceagent.Firm` [1]_ or to create a
consumer the :class:`abceagent.Household`.

For detailed documentation on:

Trading:
    see :class:`abceagent.Trade`

Logging and data creation:
    see :class:`abceagent.Database` and :doc:`simulation_results`

Messaging between agents:
    see :class:`abceagent.Messaging`.

.. autoexception:: abce.NotEnoughGoods

.. [1] or :class:`abceagent.FirmMultiTechnologies` for  complex technologies.
"""
from collections import OrderedDict


class Database:
    """ The database class """
    def __init__(self, id, agent_parameters, simulation_parameters, group, trade_logging,
                 database, check_unchecked_msgs, expiring, perishable, resource_endowment, start_round=None):
        super(Database, self).__init__(id, agent_parameters, simulation_parameters, group, trade_logging, database,
                                       check_unchecked_msgs, expiring, perishable, resource_endowment, start_round)

    def log(self, action_name, data_to_log):
        """ With log you can write the models data. Log can save variable
        states and and the working of individual functions such as production,
        consumption, give, but not trade(as its handled automatically). Sending
        a dictionary instead of several using several log statements with a
        single variable is faster.

        Args:
            'name'(string):
                the name of the current action/method the agent executes

            data_to_log:
                a variable or a dictionary with data to log in the the database

        Example::

            self.log('profit', profit)

            self.log('employment_and_rent',
                     {'employment': self['LAB'],
                     'rent': self['CAP'],
                     'composite': self.composite})

            self.log(self.produce_use_everything())

        See also:
            :meth:`~abecagent.Database.log_nested`:
                handles nested dictianaries

            :meth:`~abecagent.Database.log_change`:
                loges the change from last round

            :meth:`~abecagent.Database.observe_begin`:

        """
        if self.log_this_round:
            try:
                data_to_write = {'%s_%s' % (str(action_name), str(
                    key)): data_to_log[key] for key in data_to_log}
            except TypeError:
                data_to_write = {str(action_name): data_to_log}

            self.database_connection.put(
                ["log",
                 self.group,
                 self.id,
                 self.round,
                 data_to_write,
                 action_name])

    def log_change(self, action_name, data_to_log):
        """ This command logs the change in the variable from the round before.
        Important, use only once with the same action_name.

        Args:
            'name'(string):
                the name of the current action/method the agent executes
            data_to_log:
                a dictianary with data for the database

        Examples::

            self.log_change('profit', {'money': self['money']]})
            self.log_change('inputs',
                {'money': self.possessions(['money', 'gold', 'CAP', 'LAB')]})
        """
        if self.log_this_round:
            data_to_write = {}
            try:
                for key in data_to_log:
                    data_to_write['%s_change_%s' % (
                        action_name, key)] = (
                            data_to_log[key] -
                            self._data_to_log_1[action_name][key])
            except KeyError:
                for key in data_to_log:
                    data_to_write['%s_change_%s' %
                                  (action_name, key)] = data_to_log[key]
            data_to_write['id'] = self.id
            self.database_connection.put(
                ["log", self.group, data_to_write, str(self.round)])

            self._data_to_log_1[action_name] = data_to_log

    def observe_begin(self, action_name, data_to_observe):
        """ observe_begin and observe_end, observe the change of a variable.
        observe_begin(...), takes a list of variables to be observed.
        observe_end(...) writes the change in this variables into the log file

        you can use nested observe_begin / observe_end combinations

        Args:
            'name'(string):
                the name of the current action/method the agent executes
            data_to_log:
                a dictianary with data for the database

        Example::

            self.log('production', {'composite': self.composite,
                self.sector: self.final_product[self.sector]})

            ... different method ...

            self.log('employment_and_rent', {
                'employment': self['LAB'],
                'rent': self['CAP']})
        """
        if self.log_this_round:
            self._data_to_observe[action_name] = data_to_observe

    def observe_end(self, action_name, data_to_observe):
        """ This command puts in a database called log, whatever values you
        want values need to be delivered as a dictionary:

        Args:
            'name'(string):
                the name of the current action/method the agent executes
            data_to_log:
                a dictianary with data for the database

        Example::

            self.log('production', {'composite': self.composite,
                self.sector: self.final_product[self.sector]})

            ... different method ...

            self.log('employment_and_rent', {
                'employment': self['LAB'],
                'rent':self['CAP']})
        """
        if self.log_this_round:
            before = self._data_to_observe.pop(action_name)
            data_to_write = {}
            for key in data_to_observe:
                data_to_write['%s_delta_%s' % (action_name, key)] = \
                    data_to_observe[key] - before[key]
            data_to_write['id'] = self.id
            self.database_connection.put(
                ["log", self.group, data_to_write, str(self.round)])

    def _common_log(self, variables, possessions, functions, lengths):
        ret = OrderedDict()
        for var in variables:
            ret[var] = self.__dict__[var]
        for pos in possessions:
            ret[pos] = self._inventory[pos]
        for name, func in functions.items():
            ret[name] = func(self)
        for length in lengths:
            ret['len_' + length] = len(self.__dict__[length])
        return ret

    def _agg_log(self, variables, possessions, functions, lengths):
        if self.log_this_round:
            data_to_write = self._common_log(variables,
                                             possessions,
                                             functions,
                                             lengths)
            self.database_connection.put(["snapshot_agg",
                                          str(self.round),
                                          self.group,
                                          data_to_write])

    def _panel_log(self, variables, possessions, functions, lengths, serial):
        if self.log_this_round:
            data_to_write = self._common_log(variables,
                                             possessions,
                                             functions,
                                             lengths)
            self.database_connection.put(["log",
                                          self.group,
                                          self.id,
                                          str(self.round),
                                          data_to_write,
                                          serial])



import abce
import time


class LoggerTest(abce.Agent):
    def init(self, rounds):
        self.last_round = rounds - 1
        self.create('money', 50)
        self.create('cookies', 3)

    def one(self):
        self.log('possessions', self.possessions())
        self.log('round_log', self.round)
        pass

    def two(self):
        pass

    def three(self):
        pass

    def clean_up(self):
        pass

    def all_tests_completed(self):
        if self.round == self.last_round:
            time.sleep(0.5)
            print('Check database whether logging succeeded')



""" A graphical user interface for ABCE simulations """
import os
from os.path import join
import json
from hashlib import sha1
from collections import defaultdict
import pandas as pd
from flexx import ui
import abce
from .dockpanel import DockPanel
from .make_graphs import (make_panel_graphs,
                          make_simple_graphs,
                          make_aggregate_graphs,
                          make_histograms)
from .bokehwidget import BokehWidget
from .webtext import abcedescription
from .loadform import LoadForm


def basiclayout(Form, simulation, title, top_bar=None, story=None,
                texts=None, pages=None, truncate_rounds=0,
                histograms=None, graphs=False):
    """ Generates the basic layout of the website  """
    story = ({} if story is None else story)
    pages = ({} if pages is None else pages)
    texts = ([abcedescription] if texts is None else texts)

    class ABCE(ui.Widget):
        """ Basic layout of the website  """
        CSS = """
        h1, a {
            -webkit-margin-before: 0.20em;
            -webkit-margin-after: 0.0em;
            -webkit-margin-start: 1em;
            -webkit-margin-end: 1em;
            text-decoration: none;
        }
        ::-webkit-scrollbar {
          -webkit-appearance: none;
          width: 10px;
        }

        ::-webkit-scrollbar-thumb {
          border-radius: 5px;
          background-color: rgba(0,0,0,.5);
          -webkit-box-shadow: 0 0 1px rgba(255,255,255,.5);
        }

         """

        def init(self):
            self.graphs = {}
            self.first = self.first_repeat = True
            with ui.BoxLayout(orientation='v',
                              style="background-color: blue;"):
                with ui.HBox(flex=0):
                    ui.Label(text='<h1 style="color: white">%s</h1>' % title,
                             flex=1,
                             style="background-color: blue;")
                if top_bar is not None:
                    ui.Label(text=top_bar,
                             flex=0,
                             style="background-color: blue;")
                with DockPanel(flex=1) as self.dockpanel:
                    self.form = Form(title='Simulation',
                                     style="location: N; overflow: scroll;")
                    self.loadform = LoadForm(
                        title='Load',
                        style="location: A; overflow-y: scroll;")
                    for text in texts:
                        ui.Label(title=text.splitlines()[0],
                                 text='\n'.join(text.splitlines()[1:]),
                                 style="location: R; overflow: scroll;",
                                 wrap=True)
                    for pagetitle, page in pages:
                        ui.IFrame(url=page,
                                  title=pagetitle,
                                  style="location: A; overflow: scroll;")
                    self.progress_label = ui.Label(
                        title=' ',
                        text='Move tabs by dragging tabs; resize windows',
                        style="location: S; overflow: scroll;",
                        wrap=True)

            if graphs:
                self.display_results({'simulation_name': ''}, '')

            @self.form.connect("run_simulation")
            def run_simulation(events):
                """ Runs simulation and shows results """
                self.display_status('Running...', 'Simulation in progress')
                abce.simulation_name, parameters = (
                    hash_simulation_parameters(events))
                simulation(parameters)
                self.display_status('Success',
                                    'Simulation succeeded, generating graphs')
                self.display_results(events, abce.simulation_name)
                self.display_status('Results:', 'Click left')
                del abce.simulation_name

            @self.form.connect("repeatexecution")
            def _repeatexecution(events):
                name, parameters = hash_simulation_parameters(events)
                print("parameters:", parameters)
                parameters['Name'] = name
                pool_path = join(os.path.abspath('./result/cache'), name)

                if name not in self.graphs:
                    self.graphs[name] = load_cached(pool_path)

                self.display_status('Continuously updating',
                                    'Series of simulations in progress')
                abce.simulation_name = name
                switch_on_conditional_logging(parameters, histograms)
                simulation(parameters)
                del abce.simulation_name
                del abce.conditional_logging
                path = newest_subdirectory('./result', name)
                for filename in os.listdir(path):
                    if (filename != 'trade.csv' and
                            filename.endswith('.csv')):
                        self.graphs[name][filename] = (
                            self.graphs[name][filename]
                            .append(pd.read_csv(join(path, filename)))
                            .reset_index(drop=True))
                        if len(self.graphs[name][filename]) % 10 == 0:
                            self.graphs[name][filename].to_pickle(
                                join(pool_path, filename))
                number_obs = len(list(self.graphs[name].values())[0])
                if number_obs < 12 or number_obs % 10 == 0:
                    self.display_repeat_execution(self.graphs[name])
                self.form.emit('_repeat_execution', events)

            @self.form.connect('display_results')
            def display_results(events):  # pylint: disable=W0612
                """ Forwarder """
                self.display_results(events, events['simulation_name'])

            @self.form.connect('update_parameter_database')
            def _update_parameter_database(events):
                """ Forwarder """
                self.loadform.update(events)

            @self.loadform.connect("load")
            def _load(event):
                self.form.load_parameter(event)
                self.dockpanel.selectWidget(self.form)

        def display_status(self, title, text):
            """ displays status on bottom left
            as first tab of results panel """
            self.progress_label.title = title
            self.progress_label.text = text

        def display_results(self, events, simulation_name):  # pylint: disable=R0912
            """ Displays results of single simulation_name """
            if self.first:
                self.plot_widgets = []
            try:
                ignore_initial_rounds = int(events['ignore_initial_rounds'])
            except KeyError:
                ignore_initial_rounds = 100

            try:
                path = events['subdir']
            except KeyError:
                path = newest_subdirectory('./result', simulation_name)

            i = 0
            for filename in os.listdir(path):
                if not filename.endswith('.csv'):
                    continue

                table = pd.read_csv(path + filename).ix[truncate_rounds:]
                try:
                    rounds = max(table['round'])
                except KeyError:
                    rounds = max(table['index'])
                if ignore_initial_rounds >= rounds:
                    ignore_initial_rounds = 0
                    print('abcegui.py ignore_initial_rounds >= rounds')
                if (filename.startswith('aggregate_') or
                        filename.startswith('aggregated_')):
                    titles, plots = make_aggregate_graphs(
                        table, filename, ignore_initial_rounds)
                else:
                    try:
                        if max(table.get('id', [0])) == 0:
                            titles, plots = make_simple_graphs(
                                table, filename, ignore_initial_rounds)
                        else:
                            titles, plots = make_panel_graphs(table, filename, ignore_initial_rounds)
                    except ValueError as error:
                        print(filename, 'not displayable: ValueError', error)

                if self.first:
                    with self.dockpanel:
                        for plottitle, plot in zip(titles, plots):
                            plotwidget = BokehWidget(plot=plot,
                                                     style="location: A",
                                                     title=plottitle)
                            self.plot_widgets.append(plotwidget)
                        self.dockpanel.selectWidget(self.plot_widgets[0])
                else:
                    for plot in plots:
                        self.plot_widgets[i].plot = plot
                        i += 1

            self.first = False

        def display_repeat_execution(self, graphs):
            """ Displays histograms of repeat executions """
            if self.first_repeat:
                self.repeat_plot_widgets = []
            i = 0
            for filename, data in graphs.items():
                titles, plots = make_histograms(data, filename)

                if self.first_repeat:
                    with self.dockpanel:
                        for plottitle, plot in zip(titles, plots):
                            plotwidget = BokehWidget(plot=plot,
                                                     style="location: A",
                                                     title=plottitle)
                            self.repeat_plot_widgets.append(plotwidget)
                        self.dockpanel.selectWidget(
                            self.repeat_plot_widgets[0])
                else:
                    for plot in plots:
                        self.repeat_plot_widgets[i].plot = plot
                        i += 1
            self.first_repeat = False
    return ABCE


def newest_subdirectory(directory='.', name=''):
    """ Returns the newes subdirectory in the 'directory/name' directory """
    directory = os.path.abspath(directory)
    all_subdirs = [os.path.join(directory, name)
                   for name in os.listdir(directory)
                   if os.path.isdir(os.path.join(directory, name))]
    all_subdirs = sorted(all_subdirs, key=os.path.getmtime, reverse=True)
    for subdir in all_subdirs:
        if name in subdir:
            return subdir + '/'
    raise Exception()


def hash_simulation_parameters(events):
    """ hashes simulation parameter, after setting random seed to None """
    parameters = events['simulation_parameter']
    parameters['random_seed'] = None
    parameters['name'] = None
    parameters['Name'] = None
    name = sha1(json.dumps(parameters, sort_keys=True)
                .encode('utf-8')).hexdigest()
    return name, parameters


def load_cached(pool_path):
    """ Loads a file and creates directory if file does not exist    """
    graphs = defaultdict(pd.DataFrame)
    try:
        for filename in os.listdir(pool_path):
            graphs[filename] = pd.read_pickle(join(pool_path, filename))
    except FileNotFoundError:
        os.makedirs(pool_path)
    return graphs


def switch_on_conditional_logging(parameters, histograms):
    """ Uses abce.conditional_logging, to instruct the simulation, to log
    only at a specific point of time """
    if histograms is not None:
        abce.conditional_logging = histograms
    elif 'rounds' in parameters:
        abce.conditional_logging = [parameters['rounds'] - 1]
    elif 'histogram' in parameters:
        abce.conditional_logging = [parameters['histogram']]
    else:
        raise Exception("In @gui specify when histograms should be produced")



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# ABCE is open-source software. If you are using ABCE for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
from collections import defaultdict
from random import shuffle


class Quotation(object):
    __slots__ = ['sender_group',
                 'sender_id',
                 'receiver_group',
                 'receiver_id',
                 'good',
                 'quantity',
                 'price',
                 'buysell',
                 'id']

    def __init__(self,
                 sender_group,
                 sender_id,
                 receiver_group,
                 receiver_id,
                 good,
                 quantity,
                 price,
                 buysell,
                 id):

        self.sender_group = sender_group
        self.sender_id = sender_id
        self.receiver_group = receiver_group
        self.receiver_id = receiver_id
        self.good = good
        self.quantity = quantity
        self.price = price
        self.buysell = buysell
        self.id = id


class Quote(object):
    """ Quotes as opposed to trades are uncommitted offers. They can be made
    even if they agent can not fullfill them. With
    :meth:`~abceagent.Trade.accept_quote` and
    :meth:`~abceagent.Trade.accept_quote_partial`,
    the receiver of a quote can transform them into a trade.
    """

    def get_quotes(self, good, descending=False):
        """ self.get_quotes() returns all new quotes and removes them. The order
        is randomized.

        Args:
            good:
                the good which should be retrieved
            descending(bool,default=False):
                False for descending True for ascending by price

        Returns:
         list of quotes ordered by price

        Example::

         quotes = self.get_quotes()
        """
        ret = []
        for offer_id in list(self._quotes.keys()):
            if self._quotes[offer_id].good == good:
                ret.append(self._quotes[offer_id])
                del self._quotes[offer_id]
        ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def get_quotes_all(self, descending=False):
        """ self.get_quotes_all() returns a dictionary with all now new quotes ordered
        by the good type and removes them. The order is randomized.

        Args:
            descending(bool,default=False):
                False for descending True for ascending by price

        Returns:
            dictionary of list of quotes ordered by price. The dictionary
            itself is ordered by price.

        Example::

            quotes = self.get_quotes()
        """
        ret = defaultdict(list)

        for quote in self._quotes:
            key = self._quotes[quote].good
            ret[key].append(self._quotes[quote])
        for key in list(ret.keys()):
            shuffle(ret[key])
            ret[key].sort(key=lambda objects: objects.price,
                          reverse=descending)
        self._quotes = {}
        return ret

    def accept_quote(self, quote):
        """ makes a commited buy or sell out of the counterparties quote. For
        example, if you receive a buy quote you can accept it and a sell
        offer is send to the offering party.

        Args::
         quote: buy or sell quote that is accepted

        """
        if quote.buysell == 'qs':
            self.buy((quote.sender_group, quote.sender_id),
                     quote.good, quote.quantity, quote.price)
        else:
            self.sell((quote.sender_group, quote.sender_id),
                      quote.good, quote.quantity, quote.price)

    def accept_quote_partial(self, quote, quantity):
        """ makes a commited buy or sell out of the counterparties quote

        Args::
         quote: buy or sell quote that is accepted
         quantity: the quantity that is offered/requested
         it should be less than propsed in the quote, but this is not enforced.

        """
        if quote.buysell == 'qs':
            self.buy((quote.sender_group, quote.sender_id),
                     quote.good, quantity, quote.price)
        else:
            self.sell((quote.sender_group, quote.sender_id),
                      quote.good, quantity, quote.price)

    def quote_sell(self, receiver,
                   good=None, quantity=None, price=None):
        """ quotes a price to sell quantity of 'good' to a receiver. Use None,
        if you do not want to specify a value.

        price (money) per unit
        offers a deal without checking or committing resources

        Args:
            receiver_group:
                agent group name of the agent
            receiver_id:
                the agent's id number
            'good':
                name of the good
            quantity:
                maximum units disposed to sell at this price
            price:
                price per unit
        """
        offer = Quotation(sender_group=self.group,
                          sender_id=self.id,
                          receiver_group=receiver[0],
                          receiver_id=receiver[1],
                          good=good,
                          quantity=quantity,
                          price=price,
                          buysell='qs',
                          id=self._offer_counter())
        self._send(receiver[0], receiver[1], '_q', offer)
        return offer

    def quote_buy(self, receiver,
                  good=None, quantity=None, price=None):
        """ quotes a price to buy quantity of 'good' a receiver. Use None,
        if you do not want to specify a value.

        price (money) per unit
        offers a deal without checking or committing resources

        Args:
            receiver_group:
                agent group name of the agent
            receiver_id:
                the agent's id number
            'good':
                name of the good
            quantity:
                maximum units disposed to buy at this price
            price:
                price per unit
        """
        offer = Quotation(sender_group=self.group,
                          sender_id=self.id,
                          receiver_group=receiver[0],
                          receiver_id=receiver[1],
                          good=good,
                          quantity=quantity,
                          price=price,
                          buysell='qb',
                          id=self._offer_counter())
        self._send(receiver[0], receiver[1], '_q', offer)
        return offer



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# ABCE is open-source software. If you are using ABCE for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
""" This is the agent's facility to send and receive messages. Messages can
either be sent to an individual with :meth:`messaging.Messaging.message` or to a group with
:meth:`messaging.Messaging.message_to_group`. The receiving agent can either get all messages
with  :meth:`messaging.Messaging.get_messages_all` or messages with a specific topic with
:meth:`messaging.Messaging.get_messages`.
"""
from random import shuffle


class Message(object):
    __slots__ = ['sender', 'receiver', 'topic', 'content']

    def __init__(self, sender, receiver, topic, content):
        self.sender = sender
        self.receiver = receiver
        self.topic = topic
        self.content = content

    def __getitem__(self, item):
        return self.content[item]

    def __repr__(self):
        return "<{sender: %s; receiver: %s; topic: %s; content: %s}>" % (
            str(self.sender), str(self.receiver), self.topic, str(self.content))


class Messaging:
    def __init__(self, id, agent_parameters, simulation_parameters, group, trade_logging,
                 database, check_unchecked_msgs, expiring, perishable, resource_endowment, start_round=None):
        super(Messaging, self).__init__(id, agent_parameters, simulation_parameters, group, trade_logging, database,
                                        check_unchecked_msgs, expiring, perishable, resource_endowment, start_round)

    def send(self, receiver, topic, content):
        """ sends a message to agent. Agents receive it
        at the beginning of next round with :meth:`~abceagent.Messaging.get_messages` or
        :meth:`~abceagent.Messaging.get_messages_all`.

        Args::

            receiver:
                The name of the receiving agent a tuple (group, id).
                e.G. ('firm', 15)

            topic:
                string, with which this message can be received

            content:
                string, dictionary or class, that is send.

        Example::

            ... household_01 ...
            self.message('firm', 01, 'quote_sell', {'good':'BRD', 'quantity': 5})

            ... firm_01 - one subround later ...
            requests = self.get_messages('quote_sell')
            for req in requests:
                self.sell(req.sender, req.good, reg.quantity, self.price[req.good])

        Example2::

         self.message('firm', 01, 'm', "hello my message")

        """
        msg = Message(sender=self.name,
                      receiver=receiver,
                      topic=topic,
                      content=content)
        self._send(receiver[0], receiver[1], topic, msg)

    def get_messages(self, topic='m'):
        """ self.messages() returns all new messages send with :meth:`~abceagent.Messaging.message`
        (topic='m'). The order is randomized. self.messages(topic) returns all
        messages with a topic.

        A message is a string with the message. You can also retrieve the sender
        by `message.sender_group` and `message.sender_id` and view the topic with
        'message.topic'. (see example)

        If you are sending a float or an integer you need to access the message
        content with `message.content` instead of only `message`.

        ! if you want to recieve a **float** or an **int**, you must msg.content

        Returns a message object:
            msg.content:
                returns the message content string, int, float, ...
            msg:
                returns also the message content, but only as a string
            sender_group:
                returns the group name of the sender
            sender_id:
                returns the id of the sender
            topic:
                returns the topic

        Example::

         ... agent_01 ...
         self.messages('firm_01', 'potential_buyers', 'hello message')

         ... firm_01 - one subround later ...
         potential_buyers = get_messages('potential_buyers')
         for msg in potential_buyers:
            print('message: ', msg)
            print('message: ', msg.content)
            print('group name: ', msg.sender_group)
            print('sender id: ', msg.sender_id)
            print('topic: ', msg.topic)

        """
        try:
            shuffle(self._msgs[topic])
        except KeyError:
            self._msgs[topic] = []
        return self._msgs.pop(topic)

    def get_messages_all(self):
        """ returns all messages irregardless of the topic, in a dictionary by topic

        A message is a string with the message. You can also retrieve the sender
        by `message.sender_group` and `message.sender_id` and view the topic with
        'message.topic'. (see example)

        If you are sending a float or an integer you need to access the message
        content with `message.content` instead of only `message`.
        """
        ret = {}
        for key, messages in self._msgs.items():
            shuffle(messages)
            ret[key] = messages
        self._msgs.clear()
        return ret



from multiprocessing.managers import BaseManager
from collections import defaultdict
from copy import deepcopy


class Postbox:
    def __init__(self):
        self.boxes = defaultdict(dict)

    def receive(self, group, id):
        return self.boxes[group][id]

    def clear(self, group):
        self.boxes[group] = {}

    def copy(self, group_names):
        return {group: deepcopy(self.boxes[group]) for group in group_names}

    def send(self, group, id, message):
        try:
            self.boxes[group][id].append(message)
        except KeyError:
            self.boxes[group][id] = [message]


class PostboxManager(BaseManager):
    pass



import platform
import start_core_engine
import start_logging_test
import start_delete_create
import start_production_consumption
import start_combinable_actions
import start_return


if __name__ == '__main__':
    print("Logging test, 1 core")
    start_logging_test.main(processes=1)
    print('Iteration of logger testing with 1 core finished')

    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        print("Logging test, 4 cores")
        start_logging_test.main(processes=4)
        print('Iteration of logger testing with multiple processes finished')

    print("\nCore engine test, 1 core")
    start_core_engine.main(processes=1, rounds=50)
    print('Iteration with 1 core finished')

    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        print("Core engine test, 4 cores")
        start_core_engine.main(processes=4, rounds=50)
        print('Iteration with multiple processes finished')

    print("\nCreate/delete, 1 core")
    start_delete_create.main(processes=1, rounds=30)
    print('Iteration with 1 core finished')

    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        print("Create/delete, 4 cores")
        start_delete_create.main(processes=4, rounds=20)
        print('Iteration with multiple processes finished')

    print("PYPY and windows: core functions not tested with multi-processes")

    print("\nProduction and consumption test, 1 core")
    start_production_consumption.main(processes=1)
    print('Iteration of production and consumption testing with 1 core finished')

    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        print("Production and consumption test, 4 cores")
        start_production_consumption.main(processes=4)
        print('Iteration of production and consumption testing with multiple processes finished')

    print("\nCombinable actions test, 1 core")
    start_combinable_actions.main(processes=1, rounds=1)
    print('Iteration of combinable actions with 1 core finished')

    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        print("Combinable actions test, 4 cores")
        start_combinable_actions.main(processes=4, rounds=1)
        print('Iteration of combinable actions with multiple processes finished')

    print("\Returning and parameters test, 1 core")
    start_return.main(processes=1, rounds=5)
    print('Iteration of returning and parameters with 1 core finished')

    if (platform.system() != 'Windows' and platform.python_implementation() != 'PyPy'):
        print("Returning and parameters test, 4 cores")
        start_return.main(processes=4, rounds=5)
        print('Iteration of returning and parameters test with multiple processes finished')



from buy import Buy
# from quote_buy import QuoteBuy
from sell import Sell
from give import Give  # tests give and messaging
from logger_test import LoggerTest
from endowment import Endowment
from buyexpiringcapital import BuyExpiringCapital
from abce import Simulation
from messagea import MessageA
from messageb import MessageB


def main(processes, rounds):
    s = Simulation(processes=processes, name='unittest')
    s.declare_round_endowment(
        resource='labor_endowment', units=5, product='labor')
    s.declare_round_endowment(resource='cow', units=10,
                              product='milk')
    s.declare_perishable(good='labor')

    # s.declare_expiring('xcapital', 5)
    print('build Buy')
    buy = s.build_agents(Buy, 'buy', 1000, rounds=rounds)
    print('build Sell')
    # s.build_agents(QuoteBuy, 2)
    sell = s.build_agents(Sell, 'sell', 1000, rounds=rounds)
    print('build Give')
    give = s.build_agents(Give, 'give', 2, rounds=rounds)  # tests give and messaging
    print('build Endowment')
    endowment = s.build_agents(Endowment, 'endowment', 2, rounds=rounds, creation=0)
    # tests declare_round_endowment and declare_perishable
    print('build LoggerTest')
    loggertest = s.build_agents(
        LoggerTest, 'loggertest', 1, rounds=rounds)

    # print('build ContractSeller')
    # contractseller = s.build_agents(ContractSeller, 'contractseller', 2,
    #    rounds=rounds)
    # print('build ContractBuyer')
    # contractbuyer = s.build_agents(ContractBuyer, 'contractbuyer', 2,
    #    rounds=rounds)
    # print('build ContractSellerStop')
    # contractsellerstop = s.build_agents(ContractSellerStop,
    #    'contractsellerstop', 2, rounds=rounds)
    # print('build ContractBuyerStop')
    # contractbuyerstop = s.build_agents(ContractBuyerStop,
    #    'contractbuyerstop', 2, rounds=rounds)
    # s.build_agents(ExpiringCapital, 1)
    # s.build_agents(GiveExpiringCapital, 2)
    print('build BuyExpiringCapital')
    _ = s.build_agents(BuyExpiringCapital, 'buyexpiringcapital', 2,
                       rounds=rounds)
    print('build MessageA')
    messagea = s.build_agents(MessageA, 'messagea', 20)
    print('build MessageB')
    messageb = s.build_agents(MessageB, 'messageb', 20)

    some = buy + sell + give + loggertest
#    contractagents = (contractbuyer + contractseller
#                      + contractbuyerstop + contractsellerstop)

    for r in range(rounds):
        s.advance_round(r)
        for _ in range(5):
            buy.one()
            buy.two()
            buy.three()
            buy.clean_up()
        buy.panel_log(variables=['price'])
        for _ in range(5):
            sell.one()
            sell.two()
            sell.three()
            sell.clean_up()
        for _ in range(5):
            give.one()
            give.two()
            give.three()
            give.clean_up()
        for _ in range(5):
            loggertest.one()
            loggertest.two()
            loggertest.three()
            loggertest.clean_up()
        (messagea + messageb).sendmsg()
        (messageb + messagea).recvmsg()
        endowment.Iconsume()
        # (contractbuyer + contractbuyerstop).request_offer()
        # (contractseller + contractsellerstop).make_offer()
        # contractagents.accept_offer()
        # contractagents.deliver()
        # contractagents.pay()
        # contractagents.control()

        some.all_tests_completed()
    s.finalize()


if __name__ == '__main__':
    main(processes=1, rounds=5)
    print('Iteration with 1 core finished')
    main(processes=4, rounds=5)
    print('Iteration with multiple processes finished')



import platform
import csv
import abce
try:
    from math import isclose
except ImportError:
    def isclose(a, b):
        return a - 0.000001 < b < a + 0.000001


class Agent(abce.Agent):
    def go(self):
        self.create('money', 0.1)
        self.i = self.id
        self.r = self.round
        self.log('li', self.i)
        self.log('lr', self.r)
        self.log('l', {'i': self.i, 'r': self.r})


def compare(to_compare, path, message, processes):
    the_path = (path + '/' + to_compare)
    if platform.system() == 'Windows':  # windows compatibility
        the_path = the_path[the_path.find('/') + 1:]
    with open(to_compare, 'r') as generatedf:
        generated = {}
        for row in csv.DictReader(generatedf):
            try:
                generated[(row['round'], row['id'])] = row
            except KeyError:
                generated[row['round']] = row
    with open(the_path, 'r') as orginialf:
        orginial = {}
        for row in csv.DictReader(orginialf):
            try:
                orginial[(row['round'], row['id'])] = row
            except KeyError:
                orginial[row['round']] = row
    for row in generated:
        for key in generated[row]:
            if key != 'index':
                if generated[row][key] != orginial[row][key]:
                    assert isclose(float(generated[row][key]), float(orginial[row][key])), (
                        key, generated[row][key], orginial[row][key])
    for row in orginial:
        for key in orginial[row]:
            if key != 'index':
                if generated[row][key] != orginial[row][key]:
                    assert isclose(float(generated[row][key]), float(orginial[row][key])), (
                        key, generated[row][key], orginial[row][key])


def main(processes):
    simulation = abce.Simulation(processes=processes)

    agents = simulation.build_agents(Agent, 'agent', 10)

    for rnd in range(100):
        simulation.advance_round(rnd)
        agents.go()
        agents.agg_log(variables=['i', 'r'], goods=['money'])
        agents.panel_log(variables=['i', 'r'], goods=['money'])
    simulation.finalize()

    if platform.system() == 'Windows':
        simulation.path = simulation.path.replace('/', '\\')

    compare('aggregated_agent.csv',
            simulation.path, 'aggregated logging test\t\t',
            processes)
    compare('aggregate_agent.csv',
            simulation.path, 'aggregate logging test\t\t',
            processes)
    compare('panel_agent.csv',
            simulation.path, 'aggregate logging test mean\t',
            processes)


if __name__ == '__main__':
    main(processes=1)
    main(processes=4)



# pylint: disable=W0232, C1001, C0111, R0913, E1101, W0212
# TODO end_contract; record all payments
from __future__ import print_function
from builtins import str
from builtins import object
from abce.notenoughgoods import NotEnoughGoods
from random import shuffle
from abce.trade import get_epsilon

epsilon = get_epsilon()


class Credit(object):
    __slots__ = ['sender_group', 'sender_id', 'deliver_good_group',
                 'deliver_good_id', 'pay_group', 'pay_id', 'amount', 'interest']

    def __init__(self, sender_group, sender_id, deliver_good_group,
                 deliver_good_id, pay_group, pay_id, quantity, interest):
        self.sender_group = sender_group
        self.sender_id = sender_id
        self.deliver_good_group = deliver_good_group
        self.deliver_good_id = deliver_good_id
        self.pay_group = pay_group
        self.pay_id = pay_id
        self.quantity = quantity
        self.interest = interest
        self.id = id
        self.round = round

    def __str__(self):
        return str(('sender', self.sender_group, self.sender_id, 'deliver', self.deliver_good_group,
                    self.deliver_good_id, self.pay_group, self.pay_id, self.quantity, self.interest))


class Credit(object):
    """ This class implements a bank credit, without a due date

    """

    def request_credit(self, receiver_group, receiver_id, quantity, interest):
        """This method offers a contract to provide a good or service to the
        receiver. For a given time at a given price.

        Args:

            receiver_group:
                group to receive the good
            receiver_id:
                group to receive the good
            quantity:
                original amount to be borrowed
            interest:
                period interest
        """
        offer = Credit(sender_group=self.group,
                       sender_id=self.id,
                       deliver_good_group=self.group,
                       deliver_good_id=self.id,
                       pay_group=receiver_group,
                       pay_id=receiver_id,
                       quantity=quantity,
                       interest=interest)
        self._send(receiver_group, receiver_id, '!o', offer)
        self._contract_offers_made[offer.id] = offer
        return offer

    def get_contract_offers(self, good, descending=False):
        """ Returns all contract offers and removes them. The contract
        are ordered by price (ascending), when tied they are randomized.

        Args:
            good:
                good that underlies the contract

            descending(bool,default=False):
                False for descending True for ascending by price

        Returns:
            list of contract offers ordered by price
        """
        ret = self._contract_offers[good]
        del self._contract_offers[good]
        shuffle(ret)
        ret.sort(key=lambda objects: objects.price, reverse=descending)
        return ret

    def accept_contract(self, contract, quantity=None):
        """ Accepts the contract. The contract is completely accepted, when
        the quantity is not given. Or partially when quantity is set.

        Args:

            contract:
                the contract in question, received with get_contract_requests or
                get_contract_offers

            quantity (optional):
                the quantity that is accepted. Defaults to all.
        """
        if quantity is None:
            quantity = contract.quantity
        else:
            contract.quantity = min(contract.quantity, quantity)
            assert quantity < contract.quantity + \
                epsilon * max(quantity, contract.quantity)
            if quantity > contract.quantity:
                quantity = contract.quantity

        if contract.pay_group == self.group and contract.pay_id == self.id:
            self._contracts_pay[contract.good][contract.id] = contract
            self._send(contract.sender_group,
                       contract.sender_id, '_ac', contract)
        else:
            self._contracts_deliver[contract.good][contract.id] = contract
            self._send(contract.sender_group,
                       contract.sender_id, '_ac', contract)
        # if contract is Credit:

        return contract

    def deliver_contract(self, contract):
        """ delivers on a contract """
        assert contract.deliver_good_group == self.group and contract.deliver_good_id == self.id
        quantity = contract.quantity
        available = self._haves[contract.good]
        if quantity > available + epsilon + epsilon * max(quantity, available):
            raise NotEnoughGoods(self.name, contract.good,
                                 quantity - available)
        if quantity > available:
            quantity = available

        self._haves[contract.good] -= quantity
        self._send(contract.pay_group, contract.pay_id, '_dp', contract)

    def pay_contract(self, contract):
        """ delivers on a contract """
        assert contract.pay_group == self.group and contract.pay_id == self.id
        money = contract.quantity * contract.price
        available = self._haves['money']
        if money > available + epsilon + epsilon * max(money, available):
            raise NotEnoughGoods(self.name, 'money', money - available)
        if money > available:
            money = available

        self._haves['money'] -= money
        self._send(contract.deliver_good_group,
                   contract.deliver_good_id, '_dp', contract)

    def contracts_to_deliver(self, good):
        return list(self._contracts_deliver[good].values())

    def contracts_to_receive(self, good):
        return list(self._contracts_pay[good].values())

    def contracts_to_deliver_all(self):
        ret = {}
        for good in self._contracts_deliver:
            ret[good] = list(self._contracts_deliver[good].values())
        return request_offer

    def contracts_to_receive_all(self):
        ret = {}
        for good in self._contracts_pay:
            ret[good] = list(self._contracts_pay[good].values())
        return request_offer

    def end_contract(self, contract):
        if contract.id in self._contracts_deliver[contract.good]:
            self._send(contract.pay_group, contract.pay_id,
                       '!d', ('r', contract.good, contract.id))
            del self._contracts_deliver[contract.good][contract.id]
        elif contract.id in self._contracts_pay[contract.good]:
            self._send(contract.deliver_good_group, contract.deliver_good_id,
                       '!d', ('d', contract.good, contract.id))
            del self._contracts_pay[contract.good][contract.id]
        else:
            raise Exception("Contract not found")

    def was_paid_this_round(self, contract):
        return contract.paid[-1] == self.round

    def was_delivered_this_round(self, contract):
        return contract.delivered[-1] == self.round

    def was_paid_last_round(self, contract):
        return self.round - 1 in contract.paid

    def was_delivered_last_round(self, contract):
        return self.round - 1 in contract.delivered


def bound_zero(x):
    """ asserts that variable is above zero, where foating point imprecission is accounted for,
    and than makes sure it is above 0, without floating point imprecission """
    assert x > - \
        epsilon, '%.30f is smaller than 0 - epsilon (%.30f)' % (x, - epsilon)
    if x < 0:
        return 0
    else:
        return x



"""
Simple example:

.. UIExample:: 300

    import numpy as np
    from bokeh.plotting import figure
    from flexx import app, ui, event

    x = np.linspace(0, 6, 50)

    p1 = figure()
    p1.line(x, np.sin(x))

    p2 = figure()
    p2.line(x, np.cos(x))

    class Example(ui.Widget):

        def init(self):
            with ui.BoxPanel():
                ui.BokehWidget(plot=p1)
                ui.BokehWidget(plot=p2)

"""


import os

from flexx import event
from flexx.pyscript import window
from flexx.ui import Widget

Bokeh = None  # fool flakes


class BokehWidget(Widget):
    """ A widget that shows a Bokeh plot object.

    For Bokeh 0.12 and up. The plot's ``sizing_mode`` property is set to
    ``stretch_both`` unless it was set to something other than ``fixed``. Other
    responsive modes are 'scale_width', 'scale_height' and 'scale_both`, which
    all keep aspect ratio while being responsive in a certain direction.
    """

    CSS = """
    .flx-BokehWidget > .plotdiv {
        overflow: hidden;
    }
    """

    def init(self):

        # Handle client dependencies
        import bokeh
        dev = os.environ.get('BOKEH_RESOURCES', '') == 'relative-dev'
        modname = 'bokeh.' if dev else 'bokeh.min.'
        if not (modname + 'js') in self.session.get_used_asset_names():
            res = bokeh.resources.bokehjsdir()
            if dev:
                res = os.path.abspath(
                    os.path.join(bokeh.__file__,
                                 '..', '..', 'bokehjs', 'build'))
            for x in ('css', 'js'):
                filename = os.path.join(res, x, modname + x)
                self.session.add_global_asset(modname + x, filename)

    @event.prop
    def plot(self, plot=None):
        """ The Bokeh plot object to display. In JS, this prop
        provides the corresponding backbone model.
        """

        # The sizing_mode is fixed by default, but that's silly
        try:
            if plot.sizing_mode == 'fixed':
                plot.sizing_mode = 'stretch_both'
        except AttributeError:
            pass
        if plot is not None:
            self._plot_components(plot)
        return plot

    @event.emitter
    def _plot_components(self, plot):
        from bokeh.embed import components
        script, div = components(plot)
        script = '\n'.join(script.strip().split('\n')[1:-1])
        return dict(script=script, div=div, id=plot.ref['id'])

    class JS:

        @event.prop
        def plot(self, plot=None):
            return plot

        @event.connect('_plot_components')
        def __set_plot_components(self, *events):
            ev = events[-1]
            self.node.innerHTML = ev.div
            el = window.document.createElement('script')
            el.innerHTML = ev.script
            self.node.appendChild(el)

            def getplot():
                # Get plot from id in next event-loop iter
                self.plot = Bokeh.index[ev.id]
                canvas = self.plot.plot_canvas_view
                canvas.reset_dimensions()

            window.setTimeout(getplot, 100)

        @event.connect('size')
        def __resize_plot(self, *events):
            if self.plot and self.parent:
                    self.plot.resize()



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# ABCE is open-source software. If you are using ABCE for your research you
# are requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import threading
from collections import defaultdict
import dataset
from .online_variance import OnlineVariance
from .postprocess import to_csv


class Database(threading.Thread):
    """Separate thread that receives data from in_sok and saves it into a
    database"""
    def __init__(self, directory, in_sok, trade_log):
        threading.Thread.__init__(self)
        self.directory = directory
        self.panels = {}
        self.in_sok = in_sok
        self.data = {}
        self.trade_log = trade_log

        self.aggregation = defaultdict(lambda: defaultdict(OnlineVariance))
        self.round = 0

    def run(self):
        self.dataset_db = dataset.connect('sqlite://')
        self.dataset_db.query('PRAGMA synchronous=OFF')
        # self.dataset_db.query('PRAGMA journal_mode=OFF')
        self.dataset_db.query('PRAGMA count_changes=OFF')
        self.dataset_db.query('PRAGMA temp_store=OFF')
        self.dataset_db.query('PRAGMA default_temp_store=OFF')
        table_log = {}
        current_log = defaultdict(list)
        current_trade = []
        self.table_aggregates = {}

        if self.trade_log:
            trade_table = self.dataset_db.create_table('trade___trade',
                                                       primary_id='index')

        while True:
            try:
                msg = self.in_sok.get()
            except KeyboardInterrupt:
                print("ADD simulation.finalize() after the simulation command"
                      "to write the simulation data and AVOID BLOCKING")

                break
            except EOFError:
                break

            if msg[0] == 'snapshot_agg':
                _, round, group, data_to_write = msg
                if self.round == round:
                    for key, value in data_to_write.items():
                        self.aggregation[group][key].update(value)
                else:
                    self.make_aggregation_and_write()
                    self.round = round
                    for key, value in data_to_write.items():
                        self.aggregation[group][key].update(value)

            elif msg[0] == 'trade_log':
                for (good, seller, buyer, price), quantity in msg[1].items():
                    current_trade.append({'round': msg[2],
                                          'good': good,
                                          'seller': seller,
                                          'buyer': buyer,
                                          'price': price,
                                          'quantity': quantity})
                    if len(current_trade) == 1000:
                        trade_table.insert_many(current_trade)
                        current_trade = []

            elif msg[0] == 'log':
                _, group, id, round, data_to_write, subround_or_serial = msg
                table_name = 'panel___%s___%s' % (group, subround_or_serial)
                data_to_write['round'] = round
                data_to_write['id'] = id
                current_log[table_name].append(data_to_write)
                if len(current_log[table_name]) == 1000:
                    if table_name not in table_log:
                        table_log[table_name] = self.dataset_db.create_table(
                            table_name, primary_id='index')
                    table_log[table_name].insert_many(current_log[table_name])
                    current_log[table_name] = []

            elif msg == "close":
                break

            else:
                raise Exception(
                    "abce_db error '%s' command unknown ~87" % msg)

        for name, data in current_log.items():
            if name not in self.dataset_db:
                table_log[name] = self.dataset_db.create_table(
                    name, primary_id='index')
            table_log[name].insert_many(data)
        self.make_aggregation_and_write()
        if self.trade_log:
            trade_table.insert_many(current_trade)
        self.dataset_db.commit()
        to_csv(self.directory, self.dataset_db)

    def make_aggregation_and_write(self):
        for group, table in self.aggregation.items():
            result = {'round': self.round}
            for key, data in table.items():
                result[key + '_ttl'] = data.sum()
                result[key + '_mean'] = data.mean()
                result[key + '_std'] = data.std()
            try:
                self.table_aggregates[group].insert(result)
            except KeyError:
                self.table_aggregates[group] = self.dataset_db.create_table(
                    'aggregate___%s' % group, primary_id='index')
                self.table_aggregates[group].insert(result)
            self.aggregation[group].clear()



import abce


class Endowment(abce.Agent, abce.Household):
    def init(self, rounds, creation):
        self.last_round = rounds - 1
        self.creation = creation
        self.create('labor_endowment', 1)
        self.create('cow', 1)
        self.utility_function = self.create_cobb_douglas_utility_function({'milk': 2})

    def Iconsume(self):
        assert self['labor'] == 5, self['labor']
        assert self['milk'] == 10 + (self.round - self.creation) * (10 - 3), (10 + (
            self.round - self.creation) * (10 - 3), self['milk'], self.creation)
        milk = self['milk']
        utility = self.consume(self.utility_function, {'milk': 3})
        assert utility == 9, utility
        assert milk - 3 == self['milk'], self['milk']

    def all_tests_completed(self):
        if self.round == self.last_round and self.id == 0:
            print('Test declare_round_endowment:\t\t\tOK')
            print('Test s.declare_perishable:\t\t\tOK')
            # utility testnot exaustive!



from __future__ import division
import abce


class Victim(abce.Agent, abce.Household):
    def init(self):
        # your agent initialization goes here, not in __init__
        self.count = 1

    def am_I_dead(self):
        if self.id < self.round:
            raise Exception("should be dead %i" % self.id)



from __future__ import division
from past.builtins import cmp
from builtins import str
from builtins import zip
from builtins import range
from builtins import object
from collections import deque
from abce.notenoughgoods import NotEnoughGoods
from abce.trade import get_epsilon
epsilon = get_epsilon()


class ExpiringGood(object):
    """ A good that expires after X rounds """

    def __init__(self, duration):
        self.duration = duration
        self.time_structure = deque([0 for _ in range(duration)])  # old to new

    def _advance_round(self):
        del self.time_structure[0]
        self.time_structure.append(0)

    def __add__(self, other):
        if isinstance(other, ExpiringGood):
            self.time_structure = [
                s + o for s, o in zip(self.time_structure, other.time_structure)]
        else:
            self.time_structure[-1] += other
        return self

    def __radd__(self, other):
        return other + sum(self.time_structure)

    def __sub__(self, other):
        if isinstance(other, ExpiringGood):
            other = float(other)
        if sum(self.time_structure) < - epsilon:
            raise NotEnoughGoods
        for i in range(len(self.time_structure)):
            if other >= self.time_structure[i]:
                other -= self.time_structure[i]
                self.time_structure[i] = 0
            else:
                self.time_structure[i] -= other
                break
        return self

    def __rsub__(self, other):
        return other - sum(self.time_structure)

    def __mul__(self, other):
        return sum(self.time_structure) * other

    def __floordiv__(self, other):
        return sum(self.time_structure) // other

    def __div__(self, other):
        return sum(self.time_structure) / other

    def __mod__(self, other):
        return sum(self.time_structure) % other

    def __pow__(self, other):
        return sum(self.time_structure) ** other

    def __cmp__(self, other):
        return cmp(sum(self.time_structure), other)

    def __int__(self):
        return int(sum(self.time_structure))

    def __long__(self):
        return int(sum(self.time_structure))

    def __float__(self):
        return float(sum(self.time_structure))

    def __repr__(self):
        return str(sum(self.time_structure))

    def __get__(self, instance, owner=None):
        return sum(self.time_structure)

    def __abs__(self):
        return abs(self.__float__())



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# ABCE is open-source software. If you are using ABCE for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# pylint: disable=W0212, C0111


""" The best way to start creating a simulation is by copying the start.py
file and other files from 'abce/template' in https://github.com/AB-CE/examples.

To see how to create a simulation, read :doc:`ipython_tutorial`.


This is a minimal template for a start.py::

    from agent import Agent
    from abce import *


    simulation = Simulation(name='ABCE')
    agents = simulation.build_agents(Agent, 'agent', 2)
    for time in range(100):
        simulation.advance_round(time)
        agents.one()
        agents.two()
        agents.three()
    simulation.graphs()

Note two things are important: there must be either a

:func:`~abce.simulation.graphs` or a :func:`~abce.simulation.finalize` at the end
otherwise the simulation blocks at the end.
Furthermore, every round needs to be announced using simulation.advance_round(time),
where time is any representation of time.

"""
import datetime
import os
import time
import random
import json
import queue
import multiprocessing as mp
from collections import OrderedDict
from .db import Database
from .agent import Agent, Trade  # noqa: F401
from .group import Group
from .notenoughgoods import NotEnoughGoods  # noqa: F401
from .agents import Firm, Household  # noqa: F401
from .quote import Quote  # noqa: F401
from .contracts import Contracting  # noqa: F401
from .gui import gui, graphs  # noqa: F401
from .singleprocess import SingleProcess
from .multiprocess import MultiProcess


class Simulation(object):
    """ This is the class in which the simulation is run. Actions and agents have to
    be added. Databases and resource declarations can be added. Then run
    the simulation.

    Args:
        name:
            name of the simulation

        random_seed (optional):
            a random seed that controls the random number of the simulation

        trade_logging:
            Whether trades are logged,trade_logging can be
            'group' (fast) or 'individual' (slow) or 'off'

        processes (optional):
            The number of processes that runs in parallel. Each process hosts
            a share of the agents.
            By default, if this parameter is not specified, `processes` is all
            your logical processor cores times two, using hyper-threading when available.
            For easy debugging, set processes to one and the simulation is
            executed without parallelization.
            Sometimes it is advisable to decrease the number of processes to
            the number of logical or even physical processor cores on your
            computer.
            **For easy debugging set processes to 1, this way only one agent
            runs at a time and only one error message is displayed**

        check_unchecked_msgs:
            check every round that all messages have been received with get_massages or get_offers.

        Example::

            simulation = Simulation(name='ABCE',
                                    trade_logging='individual',
                                    processes=None)


    Example for a simulation::

        num_firms = 5
        num_households = 2000

        w = Simulation(name='ABCE',
                       trade_logging='individual',
                       processes=None)

        w.declare_round_endowment(resource='labor_endowment',
                                  productivity=1,
                                  product='labor')

        w.panel('firm', command='after_sales_before_consumption')

        firms = w.build_agents(Firm, 'firm', num_firms)
        households = w.build_agents(Household, 'household', num_households)

        all = firms + households

        for r in range(100):
            self.advance_round(r)
            households.recieve_connections()
            households.offer_capital()
            firms.buy_capital()
            firms.production()
            if r == 250:
                centralbank.intervention()
            households.buy_product()
            all.after_sales_before_consumption()
            households.consume()

        w.finalize()
        w.graphs()
    """

    def __init__(self, name='abce', random_seed=None,
                 trade_logging='off', processes=1, check_unchecked_msgs=False):
        """
        """
        try:
            name = simulation_name  # noqa: F821
        except NameError:
            pass

        self.check_unchecked_msgs = check_unchecked_msgs

        self.agents_created = False
        self._messages = {}
        self._resource_command_group = {}
        self._db_commands = {}
        self.num_agents = 0
        self._build_first_run = True
        self.resource_endowment = []
        self.perishable = []
        self.expiring = []

        os.makedirs(os.path.abspath('.') + '/result/', exist_ok=True)

        self.path = (os.path.abspath('.') + '/result/' + name + '_' +
                     datetime.datetime.now().strftime("%Y-%m-%d_%H-%M"))
        """ the path variable contains the path to the simulation outcomes
        it can be used to generate your own graphs as all resulting
        csv files are there.
        """
        while True:
            try:
                os.makedirs(self.path)
                break
            except OSError:
                self.path += 'I'

        self.trade_logging_mode = trade_logging
        if self.trade_logging_mode not in ['individual', 'group', 'off']:
            Exception("trade_logging can be "
                      "'group' (fast) or 'individual' (slow) or 'off'"
                      ">" + self.trade_logging_mode + "< not accepted")

        self.processes = mp.cpu_count() * 2 if processes is None else processes

        if processes == 1:
            self._processorgroup = SingleProcess()
            self.database_queue = queue.Queue()
        else:
            self._processorgroup = MultiProcess(processes)
            manager = mp.Manager()
            self.database_queue = manager.Queue()

        self.messagess = {}

        self._db = Database(
            self.path,
            self.database_queue,
            trade_log=self.trade_logging_mode != 'off')
        self._db_started = False

        if random_seed is None or random_seed == 0:
            random_seed = time.time()
        random.seed(random_seed)

        self.sim_parameters = OrderedDict(
            {'name': name, 'random_seed': random_seed})
        self.clock = time.time()
        self.database = self
        self.time = None
        """Returns the current time set with simulation.advance_round(time)"""
        self._groups = {}

    def declare_round_endowment(self, resource, units,
                                product):
        """ At the beginning of very round the agent gets 'units' units
        of good 'product' for every 'resource' he possesses.

        Round endowments are group specific, that means when
        somebody except the specified group holds them they do not produce.

        Args::

            resource:
                The good that you have to hold to get the other

            units:
                the multiplier to get the produced good

            product:
                the good that is produced if you hold the first good

            groups:
                a list of agent groups, which gain the second good,
                if they hold the first one

        Example::

            A farmer gets a ton of harvest for every acre:

            w.declare_round_endowment(resource='land',
                                      units=1000,
                                      product='wheat')
        """
        if self.agents_created:
            raise Exception(
                "WARNING: declare_round_endowment(...)"
                " must be called before the agents are build")
        self.resource_endowment.append(
            (resource, units, product))

    def declare_perishable(self, good):
        """ This good only lasts one round and then disappears. For example
        labor, if the labor is not used today today's labor is lost.
        In combination with resource this is useful to model labor or capital.

        In the example below a worker has an endowment of labor and capital.
        Every round he can sell his labor service and rent his capital. If
        he does not the labor service for this round and the rent is lost.

        Args::

         good:
            the good that perishes

         Example::

             w.declare_perishable(good='LAB')
             w.declare_perishable(good='CAP')

        """
        if self.agents_created:
            raise Exception(
                "WARNING: declare_perishable(...) must be called before "
                "the agents are build")
        self.perishable.append(good)

    def declare_expiring(self, good, duration):
        """ This type of good lasts for several rounds, but eventually
        expires. For example computers would last for several years and than
        become obsolete.

        Args:

            good:
                the good, which expires
            duration:
                the duration before the good expires
        """
        if self.agents_created:
            raise Exception(
                "WARNING: declare_expiring(...) must be called "
                "before the agents are build")
        self.expiring.append((good, duration))

    def declare_service(self, human_or_other_resource,
                        units, service):
        """ When the agent holds the human_or_other_resource,
        he gets 'units' of service every round
        the service can be used only with in this round.

        Args::

            human_or_other_resource:
                the good that needs to be in possessions to create the other
                good 'self.create('adult', 2)'
            units:
                how many units of the service is available
            service:
                the service that is created
            groups:
                a list of agent groups that can create the service

        Example::

            For example if a household has two adult family members, it gets
            16 hours of work

            w.declare_service('adult', 8, 'work')
        """
        self.declare_round_endowment(
            human_or_other_resource, units, service)
        self.declare_perishable(service)

    def advance_round(self, time):
        if not self._db_started:
            self._db.start()
            self._db_started = True
        self.time = time
        print("\rRound" + str(time))
        self._processorgroup.advance_round(time)

    def __del__(self):
        self.finalize()

    def finalize(self):
        """ simulation.finalize() must be run after each simulation. It will
        write all data to disk

        Example::

            simulation = Simulation(...)
            ...
            for r in range(100):
                simulation.advance_round(r)
                agents.do_something()
                ...

            simulation.finalize()
        """
        if self._db_started:
            self._db_started = False
            print('')
            print(str("time only simulation %6.2f" %
                  (time.time() - self.clock)))

            self.database_queue.put('close')

            while self._db.is_alive():
                time.sleep(0.05)

            try:
                self.pool.close()
                self.pool.join()
            except AttributeError:
                pass

            print(str("time with data %6.2f" %
                      (time.time() - self.clock)))
            self._write_description_file()
            self._displaydescribtion()

    def build_agents(self, AgentClass, group_name,
                     number=None,
                     agent_parameters=None,
                     **parameters):
        """ This method creates agents.

        Args:

            AgentClass:
                is the name of the AgentClass that you imported

            group_name:
                the name of the group, as it will be used in the action list
                and transactions. Should generally be lowercase of the
                AgentClass.

            number:
                number of agents to be created.

            agent_parameters:
                a list of dictionaries, where each agent gets one dictionary.
                The number of agents is the length of the list

            any other parameters:
                are directly passed to the agent

        Example::

         firms = simulation.build_agents(Firm, 'firm',
             number=simulation_parameters['num_firms'])
         banks = simulation.build_agents(Bank, 'bank',
                                         agent_parameters=[{'name': 'UBS'},
                                         {'name': 'amex'},{'name': 'chase'}
                                         **simulation_parameters,
                                         loanable=True)

         centralbanks = simulation.build_agents(CentralBank, 'centralbank',
                                                number=1,
                                                rounds=num_rounds)
        """
        assert number is None or agent_parameters is None, \
            'either set number or agent_parameters in build_agents'
        assert group_name.isidentifier()

        if agent_parameters is None:
            agent_parameters = {}
        if parameters is None:
            parameters = {}
        if number is not None:
            agent_parameters = [{} for _ in range(number)]

        self.sim_parameters.update(parameters)

        group = Group(self, self._processorgroup, [group_name], [AgentClass],
                      agent_arguments={'group': group_name,
                                       'trade_logging': self.trade_logging_mode,
                                       'database': self.database_queue,
                                       'check_unchecked_msgs': self.check_unchecked_msgs,
                                       'expiring': self.expiring,
                                       'perishable': self.perishable,
                                       'resource_endowment': self.resource_endowment})
        group.create_agents(agent_parameters=agent_parameters, **parameters)
        self.agents_created = True
        self._groups[group_name] = group
        self.messagess[group_name] = []
        return group

    def create_agents(self, AgentClass, group_name, simulation_parameters=None, agent_parameters=None, number=1):
        """ Creates an additional agent in an existing group during the simulation. If agents
        have been deleted, their id's are reduced.

        Args:

            AgentClass:
                the class of agent to create.
                (can be the same class as the creating agent)

            'group_name':
                the name of the group the agent should belong to. This is the
                group name string e.G. :code:`'firm'`, not the group variable e.G.
                :code:`firms` in :code:`firms = simulation.build_agents(...)`

            simulation_parameters:
                a dictionary of parameters

            agent_parameters:
                List of a dictionary of parameters

            number:
                if no agent_parameters list is given the number of agents to be created can be specified

        Returns:
           id of new agent.

        Example::

            self.create_agent(BeerFirm, 'beerfirm',
                              parameters=self.parameters,
                              agent_parameters={'creation': self.time})
        """
        if simulation_parameters is None:
            simulation_parameters = {}
        if agent_parameters is None:
            agent_parameters = [{}] * number
        group = self._groups[group_name]
        id = group.create_agents(simulation_parameters=simulation_parameters,
                                 agent_parameters=agent_parameters)
        return id

    def create_agent(self, AgentClass, group_name, simulation_parameters=None, agent_parameters=None):
        raise Exception("create_agent is depreciated for create_agents")

    def delete_agent(self, *ang):
        raise Exception("delete_agent is depreciated for create_agents")

    def delete_agents(self, group, ids):
        """ This deletes a group of agents. The model has to make sure that other
        agents are notified of the death of agents in order to stop them from corresponding
        with this agent. Note that if you create new agents
        after deleting agents the ID's of the deleted agents are reused.

        Args:
            group:
                group of the agent

            ids:
                a list of ids of the agents to be deleted in that group
        """
        group = self._groups[group]
        group.delete_agents(ids)

    def _write_description_file(self):
        description = open(os.path.abspath(
            self.path + '/description.txt'), 'w')
        description.write(json.dumps(self.sim_parameters,
                                     indent=4,
                                     skipkeys=True,
                                     default=lambda x: 'not_serializeable'))

    def _displaydescribtion(self):
        description = open(self.path + '/description.txt', 'r')
        print(description.read())

    def graphs(self):
        """ after the simulation is run, graphs() shows graphs of all data
        collected in the simulation. Shows the same output as the @gui
        decorator shows.


        Example::

            simulation = Simulation(...)
            for r in range(100):
                simulation.advance_round(r)
                agents.do_something()
                ...

            simulation.graphs()
        """
        self.finalize()
        graphs(self.sim_parameters)



import abce


class ContractSeller(abce.Agent, abce.Contracting):
    def init(self, rounds):
        self.last_round = rounds - 1
        if self.id == 0:
            self.create('labor_endowment', 1)

    def make_offer(self):
        if self.id == 0:
            if self.round % 10 == 0:
                self.given_contract = self.offer_good_contract('contractseller', 1,
                                                               'labor',
                                                               quantity=5,
                                                               price=10,
                                                               duration=9)

    def accept_offer(self):
        if self.id == 1:
            contracts = self.get_contract_offers('labor')
            for contract in contracts:
                self.accepted_contract = self.accept_contract(contract)

    def deliver(self):
        for contract in self.contracts_to_deliver('labor'):
            self.deliver_contract(contract)
            assert self['labor'] == 0

    def pay(self):
        for contract in self.contracts_to_receive('labor'):
            if self.was_delivered_this_round(contract):
                self.create('money', 50)
                self.pay_contract(contract)
                assert self['money'] == 0

    def control(self):
        if self.id == 0:
            assert self.was_paid_this_round(
                self.given_contract), self._contracts_payed
            assert self['labor'] == 0, (self.id, self['labor'])
            assert self['money'] == 50, self['money']
            self.destroy('money')
        else:
            assert self.was_delivered_this_round(
                self.accepted_contract), self.contracts_to_receive('labor')
            assert self['labor'] == 5, self['labor']
            assert self['money'] == 0, self['money']

    def clean_up(self):
        pass

    def all_tests_completed(self):
        if self.round == self.last_round and self.id == 0:
            print('Test make_offer         \t\t\tOK')



abcedescription = """ ABCE is a Python based modeling platform for economic simulations. For simulations of trade, production and consumption, ABCE comes with standard functions that implement these kinds of interactions and actions. The modeler only has to implement the logic and decisions of an agent; ABCE takes care of all exchange of goods and production and consumption.

One special feature of ABCE is that goods have the physical properties of goods in reality. In other words if agent A gives a good to agent B, then - unlike information - agent B receives the good and agent B does not have the good anymore. That means that agents can trade, produce or consume a good. The ownership and transformations (production or consumption) of goods are automatically handled by the platform.

ABCE models are programmed in standard Python, stock functions of agents are inherited from archetype classes (Agent, Firm or Household). The only not-so-standard Python is that agents are executed in parallel by the Simulation class (in start.py).

ABCE allows the modeler to program agents as ordinary Python class-objects, but run the simulation on a multi-core/processor computer. It takes no effort or intervention from the modeller to run the simulation on a multi-core processor production, consumption, trade, communication and similar functions are automatically handled by the platform. The modeler only needs to instruct ABCE, which automatically executes the specific functions.

ABCE is a scheduler [1] and a set of agent classes. According to the schedule the simulation class calls - each sub-round - agents to execute some actions. Each agent executes these actions using some of the build-in functions, such as trade, production and consumption of ABCE. The agents can use the full set of commands of the Python general purpose language.

The audience of ABCE are economists that want to model agent-based models of trade and production. It is especially geared towards simulations that are similar to standard economic models like general or partial equilibrium models [2]. What is more ABCE is especially designed to make writing the simulation and the execution fast. Therefore models can be developed in an interlinked process of running and rewriting the simulation.

ABCE uses Python - a language that is especially beginner friendly, but also easy to learn for people who already know object oriented programming languages such as Java, C++ or even MATLAB. Python allows simple, but fully functional, programming for economists. What is more Python is readable even for non Python programmers.
<h2> <a href="http://abce.readthedocs.io/en/master/">Find the Documentation at readthedocs</a> </h2>"""



# Copyright 2012 Davoud Taghawi-Nejad
#
# Module Author: Davoud Taghawi-Nejad
#
# ABCE is open-source software. If you are using ABCE for your research you are
# requested the quote the use of this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License and quotation of the
# author. You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""
The :py:class:`abce.Agent` class is the basic class for creating your agents.
It automatically handles the possession of goods of an agent. In order to
produce/transforme goods you also need to subclass the :py:class:`abce.Firm` or
to create a consumer the :py:class:`abce.Household`.

For detailed documentation on:

Trading, see :doc:`Trade`

Logging and data creation, see :doc:`Database`.

Messaging between agents, see :doc:`Messaging`.
"""
from collections import defaultdict
from pprint import pprint
import abce
from .database import Database
from .trade import Trade
from .messaging import Messaging
from .goods import Goods


class DummyContracts:
    def _advance_round(self, round):
        pass


class Agent(Database, Trade, Messaging, Goods):
    """ Every agent has to inherit this class. It connects the agent to the
    simulation and to other agent. The :class:`abce.Trade`,
    :class:`abce.Database` and :class:`abce.Messaging` classes are included.
    An agent can also inheriting from :class:`abce.Firm`,
    :class:`abce.FirmMultiTechnologies` or :class:`abce.Household` classes.

    Every method can return parameters to the simulation.

    For example::

        class Household(abce.Agent, abce.Household):
            def init(self, simulation_parameters, agent_parameters):
                self.num_firms = simulation_parameters['num_firms']
                self.type = agent_parameters['type']
                ...

            def selling(self):
                for i in range(self.num_firms):
                    self.sell('firm', i, 'good', quantity=1, price=1)

            ...
            def return_quantity_of_good(self):
                return['good']


        ...

        simulation = Simulation()
        households = Simulation.build_agents(household, 'household',
                                             parameters={...},
                                             agent_parameters=[{'type': 'a'},
                                                               {'type': 'b'}])
        for r in range(10):
            simulation.advance_round(r)
            households.selling()
            print(households.return_quantity_of_good())



    """
    def __init__(self, id, agent_parameters, simulation_parameters, group, trade_logging,
                 database, check_unchecked_msgs, expiring, perishable, resource_endowment, start_round=None):
        """ Do not overwrite __init__ instead use a method called init instead.
        init is called whenever the agent are build.
        """
        super(Agent, self).__init__(id, agent_parameters, simulation_parameters, group, trade_logging,
                                    database, check_unchecked_msgs, expiring, perishable, resource_endowment,
                                    start_round)
        self.id = id
        """ self.id returns the agents id READ ONLY"""
        self.name = (group, id)
        """ self.name returns the agents name, which is the group name and the
        id
        """
        self.name_without_colon = '%s_%i' % (group, id)
        self.group = group
        """ self.group returns the agents group or type READ ONLY! """
        # TODO should be group_address(group), but it would not work
        # when fired manual + ':' and manual group_address need to be removed
        self.database_connection = database

        self.trade_logging = {'individual': 1,
                              'group': 2, 'off': 0}[trade_logging]

        self._out = []
        # TODO make defaultdict; delete all key errors regarding self._inventory as
        # defaultdict, does not have missing keys
        self._msgs = {}

        self._data_to_observe = {}
        self._data_to_log_1 = {}
        self.round = start_round
        """ self.round is depreciated"""
        self.time = start_round
        """ self.time, contains the time set with simulation.advance_round(time)
            you can set time to anything you want an integer or
            (12, 30, 21, 09, 1979) or 'monday' """
        self._resources = []
        self.variables_to_track_panel = []
        self.variables_to_track_aggregate = []
        self.inbox = []

        try:
            self._add_contracts_list()
        except AttributeError:
            self.contracts = DummyContracts()

        if hasattr(abce, 'conditional_logging'):
            self.conditional_logging = True
            self.log_rounds = abce.conditional_logging
        else:
            self.conditional_logging = False

        self.log_this_round = True

        self._check_every_round_for_lost_messages = check_unchecked_msgs

        for good, duration in expiring:
            self._declare_expiring(good, duration)

        for good in perishable:
            self._register_perish(good)

        for resource, units, product in resource_endowment:
            self._register_resource(resource, units, product)

    def init(self):
        """ This method is called when the agents are build.
        It can be overwritten by the user, to initialize the agents.
        Parameters are the parameters given to
        :py:meth:`abce.Simulation.build_agents`.

        Example::

            class Student(abce.Agent):
                def init(self, rounds, age, lazy, school_size):
                    self.rounds = rounds
                    self.age = age
                    self.lazy = lazy
                    self.school_size = school_size

                def say(self):
                    print('I am', self.age ' years old and go to a school
                    that is ', self.school_size')


            def main():
                sim = Simulation()
                students = sim.build_agents(Student, 'student',
                                            agent_parameters=[{'age': 12, lazy: True},
                                                              {'age': 12, lazy: True},
                                                              {'age': 13, lazy: False},
                                                              {'age': 14, lazy: True}],
                                            rounds=50,
                                            school_size=990)

        """
        print("Warning: agent %s has no init function" % self.group)

    def _check_for_lost_messages(self):
        for offer in list(self.given_offers.values()):
            if offer.made < self.round:
                print("in agent %s this offers have not been retrieved:" %
                      self.name_without_colon)
                for offer in list(self.given_offers.values()):
                    if offer.made < self.round:
                        print(offer.__repr__())
                raise Exception('%s_%i: There are offers have been made before'
                                'last round and not been retrieved in this'
                                'round get_offer(.)' % (self.group, self.id))

        if sum([len(offers) for offers in list(self._open_offers_buy.values())]):
            pprint(dict(self._open_offers_buy))
            raise Exception('%s_%i: There are offers an agent send that have '
                            'not been retrieved in this round get_offer(.)' %
                            (self.group, self.id))

        if sum([len(offers) for offers in list(self._open_offers_sell.values())]):
            pprint(dict(self._open_offers_sell))
            raise Exception('%s_%i: There are offers an agent send that have '
                            'not been retrieved in this round get_offer(.)' %
                            (self.group, self.id))

        if sum([len(offers) for offers in list(self._msgs.values())]):
            pprint(dict(self._msgs))
            raise Exception('(%s, %i): There are messages an agent send that '
                            'have not been retrieved in this round '
                            'get_messages(.)' % (self.group, self.id))

    def _advance_round(self, time):
        super()._advance_round(time)
        self._inventory._advance_round()
        self.contracts._advance_round(self.round)

        if self._check_every_round_for_lost_messages:
            self._check_for_lost_messages()

        for ingredient, units, product in self._resources:
            self._inventory.create(product, self[ingredient] * units)

        self.round = time
        self.time = time

        if self.conditional_logging:
            if self.round in self.log_rounds:
                print("***", self.round)
                self.log_this_round = True
            else:
                self.log_this_round = False

    def _execute(self, command, args, kwargs):
        self._clearing__end_of_subround(self.inbox)
        self.inbox.clear()
        self._begin_subround()
        ret = getattr(self, command)(*args, **kwargs)
        self._end_subround()
        self._reject_polled_but_not_accepted_offers()
        return ret

    def _post_messages(self, agents):
        for group, id, envelope in self._out:
            agents[group][id].inbox.append(envelope)
        self._out.clear()

    def _post_messages_multiprocessing(self, num_processes):
        out = self._out
        self._out = defaultdict(list)
        return out

    def _begin_subround(self):
        """ Overwrite this to make ABCE plugins, that need to do
        something at the beginning of every subround """
        pass

    def _end_subround(self):
        """ Overwrite this to make ABCE plugins, that need to do
        something at the beginning of every subround """
        pass

    def _send(self, receiver_group, receiver_id, typ, msg):
        """ sends a message to 'receiver_group', 'receiver_id'
        The agents receives it at the begin of each subround.
        """
        self._out.append(
            (receiver_group, receiver_id, (typ, msg)))

    def _send_multiprocessing(self, receiver_group, receiver_id, typ, msg):
        """ Is used to overwrite _send in multiprocessing mode.
        Requires that self._out is overwritten with a defaultdict(list) """
        self._out[receiver_id % self._processes].append(
            (receiver_group, receiver_id, (typ, msg)))

    def __del__(self):
        self._check_for_lost_messages()




import os
import platform
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

cmdclass = {}
ext_modules = []

install_requires = ['networkx >= 1.9.1',
                    'flexx >= 0.4.1',
                    'future',
                    'dataset == 0.8']


readthedocs = os.environ.get('READTHEDOCS') == 'True'

if not readthedocs:
    if not platform.python_implementation() == "PyPy":
        install_requires += ['numpy >= 1.10.2p',
                             'pandas >= 0.17.1',
                             'bokeh']


version = '0.9.4b0'


setup(name='abcEconomics',
      version=version,
      author='Davoud Taghawi-Nejad',
      author_email='Davoud@Taghawi-Nejad.de',
      description='Agent-Based Complete Economy modelling platform',
      url='https://github.com/AB-CE/abce.git',
      package_dir={'abce': 'abce',
                   'abce.gui': 'abce/gui',
                   'abce.agents': 'abce/agents',
                   'abce.contracts': 'abce/contracts'},
      packages=['abce'],
      long_description=open('README.rst').read(),
      install_requires=install_requires,
      include_package_data=True,
      ext_modules=ext_modules,
      use_2to3=True,
      cmdclass=cmdclass)



""" Copyright 2012 Davoud Taghawi-Nejad

 Module Author: Davoud Taghawi-Nejad

 ABCE is open-source software. If you are using ABCE for your research you are
 requested the quote the use of this software.

 Licensed under the Apache License, Version 2.0 (the "License"); you may not
 use this file except in compliance with the License and quotation of the
 author. You may obtain a copy of the License at
       http://www.apache.org/licenses/LICENSE-2.0
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 License for the specific language governing permissions and limitations under
 the License.
"""
# pylint: disable=W0212, C0111
from collections import ChainMap


class SingleProcess(object):
    """ This is a container for all agents. It exists only to allow for multiprocessing with MultiProcess.
    """
    def __init__(self):
        self.agents = {}

    def new_group(self, group):
        """ Creates a new group. """
        self.agents[group] = []

    def insert_or_append(self, group, ids, free_ids, Agent, simulation_parameters, agent_parameters, agent_arguments):
        """appends an agent to a group """
        if isinstance(agent_parameters, int):
            agent_parameters = ([] for _ in range(agent_parameters))

        for ap in agent_parameters:
            if free_ids:
                id = free_ids.popleft()
                ids[id] = id
            else:
                id = len(ids)
                ids.append(id)
            agent = Agent(id, simulation_parameters, ap, **agent_arguments)
            agent.init(**ChainMap(simulation_parameters, ap))
            try:
                self.agents[group][id] = agent
            except IndexError:
                self.agents[group].append(agent)
        return ids

    def delete_agents(self, group, ids):
        for id in ids:
            self.agents[group][id] = None

    def do(self, groups, ids, command, args, kwargs):
        self.rets = []
        for group, iss in zip(groups, ids):
            for i in iss:
                if i is not None:
                    ret = self.agents[group][i]._execute(command, args, kwargs)
                    self.rets.append(ret)

    def post_messages(self, groups, ids):
        for group, iss in zip(groups, ids):
            for i in iss:
                if i is not None:
                    self.agents[group][i]._post_messages(self.agents)
        return self.rets

    def advance_round(self, time):
        for agents in self.agents.values():
            for agent in agents:
                if agent is not None:
                    agent._advance_round(time)

    def group_names(self):
        return self.agents.keys()



"""Internal helper package that creates graphs for gui and
simulation.graphs()"""
import random
import pandas as pd
from bokeh.plotting import figure
from bokeh.models import Range1d, LinearAxis, Span, Label
from bokeh.charts import Histogram
from bokeh.layouts import row


COLORS = ["red", "blue", "green", "black", "purple", "pink", "yellow",
          "orange", "pink", "Brown", "Cyan", "Crimson", "DarkOrange",
          "DarkSeaGreen", "DarkCyan", "DarkBlue", "DarkViolet", "Silver",
          "#0FCFC0", "#9CDED6", "#D5EAE7", "#F3E1EB", "#F6C4E1", "#F79CD4"]

TOOLS = "reset, pan, zoom_in, zoom_out, box_zoom, save, crosshair, hover"


def make_title(table_name, col):
    """makes a title out of table_name and column, cleaning it up"""
    return (table_name.replace('aggregate_', '')
            .replace('panel_', '')
            .replace('.csv', '')
            .replace('_log_', '')
            .replace('log_', '') +
            ' ' +
            col.replace('_ttl', ''))


def abce_figure(title, y_range=None):
    return figure(title=title, width=12, height=6,
                  sizing_mode='stretch_both',
                  output_backend='webgl',
                  toolbar_location='below', tools=TOOLS, y_range=y_range)


def make_aggregate_graphs(data, filename, ignore_initial_rounds):
    """Make timeseries graphs from aggregate or aggregate_ files, which contain
    _ttl, _mean and _std suffixes, to denote total, mean and
    standard deviation columns """
    data = clean_nans(data)
    print('make_aggregate_graphs', filename)
    # all columns exist 3 times, with _ttl, _mean and _std suffix
    # columns contains each type only once:
    columns = [col.replace('_ttl', '')
               for col in data.columns if col.endswith('_ttl')]
    index = data['round']
    titles = []
    plots = []
    for col in columns:
        if col == 'index':
            continue
        title = make_title(filename, col)
        plot = abce_figure(title)
        plot.yaxis.visible = False
        plot.legend.orientation = "top_left"

        try:
            plot.extra_y_ranges['std'] = y_range(col, 'std', data,
                                                 ignore_initial_rounds)

            plot.line(index, data[col + '_std'], legend='std', line_width=2,
                      line_color='red', y_range_name="std")
            plot.add_layout(LinearAxis(y_range_name="std"), 'right')
        except KeyError:
            pass

        plot.extra_y_ranges['ttl'] = y_range(col, 'ttl', data,
                                             ignore_initial_rounds)

        plot.line(index, data[col + '_ttl'], legend='mean/total', line_width=2,
                  line_color='blue', y_range_name="ttl")
        plot.add_layout(LinearAxis(y_range_name="ttl"), 'left')
        try:
            plot.extra_y_ranges['mean'] = y_range(col, 'mean', data,
                                                  ignore_initial_rounds)
            plot.add_layout(LinearAxis(y_range_name="mean"), 'left')
        except KeyError:
            pass
        titles.append(title + ' (agg)')
        plots.append(plot)
    return titles, plots


def make_simple_graphs(data, filename, ignore_initial_rounds):
    """Make timeseries graphs from datafile, which has 'round' as
    index"""
    data = clean_nans(data)
    print('make_simple_graphs', filename)
    index = data['round']
    titles = []
    plots = []
    for col in data.columns:
        title = make_title(filename, col)
        if col not in ['round', 'id', 'index']:
            plot = abce_figure(title)
            plot.yaxis.visible = False
            plot.legend.orientation = "top_left"
            plot.extra_y_ranges['ttl'] = y_range(col, '', data,
                                                 ignore_initial_rounds)

            plot.legend.orientation = "top_left"
            plot.line(index, data[col], legend=col, line_width=2,
                      line_color='blue', y_range_name="ttl")
            plot.add_layout(LinearAxis(y_range_name="ttl"), 'left')
            titles.append(title)
            plots.append(plot)

    return titles, plots


def make_histograms(data, filename):
    """Creates histograms from data. If several 'id's are present,
    several histograms for each id are created in the same plot """
    data = clean_nans(data)
    print('make_histograms', filename)
    titles = []
    plots = []
    num_graphs = 1
    if 'id' in data.columns:
        num_graphs = min(len(set(data['id'])), 6)
    else:
        data['id'] = 0
    for col in data.columns:
        if (col not in ['round', 'id', 'index', 'index_ttl'] and
                not col.endswith('_mean') and
                not col.endswith('_std')):
            title = make_title(filename, col)
            tplot = []
            for i in range(num_graphs):
                plot = Histogram(data[data['id'] == i][col],
                                 width=1200, height=600,
                                 sizing_mode='stretch_both')
                mean = data[data['id'] == i][col].mean()
                line = Span(location=mean,
                            dimension='height', line_color='blue',
                            line_width=1)
                plot.add_layout(line)
                label = Label(x=mean, y=0, text='mean: %f' % mean)
                plot.add_layout(label)
                tplot.append(plot)
            titles.append(title + ' (hist)')
            plots.append(row(tplot, sizing_mode='stretch_both'))
    return titles, plots


def make_panel_graphs(data, filename, ignore_initial_rounds):
    """ Creates panel graphs from data with 'round' and 'id' picks no more than 20
    samples to display"""
    data = clean_nans(data)
    print('make_panel_graphs', filename)
    num_individuals = max(data['id'])
    if num_individuals > 20:
        individuals = sorted(random.sample(range(max(data['id'])), 20))
    else:
        individuals = range(max(data['id']) + 1)
    data = data[data['id'].isin(individuals)]
    titles = []
    plots = []
    for col in data.columns:
        if col not in ['round', 'id', 'index']:
            y_range = (min(data[col][ignore_initial_rounds * len(individuals):]), max(data[col][ignore_initial_rounds * len(individuals):]))
            if y_range[0] == y_range[1]:
                y_range = (-1, 1)
            title = make_title(filename, col)
            plot = abce_figure(title, y_range=y_range)
            plot.legend.orientation = "top_left"
            for i, id in enumerate(individuals):
                index = data['round'][data['id'] == id]
                series = data[col][data['id'] == id]
                plot.line(index, series, legend=str(id),
                          line_width=2, line_color=COLORS[i])
            titles.append(title + ' (panel)')
            plots.append(plot)
    return titles, plots


def clean_nans(data):
    data = data.where((pd.notnull(data)), None)
    data.dropna(1, how='all', inplace=True)
    return data


def y_range(column, suffix, data, ignore_initial_rounds):
    """ returns a range that includes min and max y
    for values where x is above ignore_initial_rounds
    """
    if suffix != "":
        column = column + '_' + suffix

    relevant_subset = data[column].ix[ignore_initial_rounds:]
    if relevant_subset.min() == relevant_subset.max():
        return Range1d(relevant_subset.min(skipna=True) - 1,
                       relevant_subset.max(skipna=True) + 1)
    return Range1d(relevant_subset.min(skipna=True),
                   relevant_subset.max(skipna=True))



""" ABCE can be started with a gui or provide visual data output """
import os
import sys
import platform
import json
from subprocess import call
import abce
import dataset
import flexx
from flexx import app
from .webtext import abcedescription
try:
    from .basiclayout import basiclayout
    from .form import form
    IMPORTERROR = False
except ImportError:
    IMPORTERROR = True


def gui(parameter_mask, names=None, header=None, story=None,
        title="Agent-Based Computational Economics",
        texts=None, pages=None, histograms=None,
        serve=False, runtime='browser-X', truncate_rounds=0,
        hostname='0.0.0.0', port=80, pypy=None):
    """ gui is a decorator that can be used to add a graphical user interface
    to your simulation.

    Args:

        parameter_mask:
            a dictionary with the parameter name as key and an example value
            as value. Instead of the example value you can also put a tuple:
            (min, default, max)

            parameters can be:
                - float:
                    {'exponent': (0.0, 0.5, 1.1)}

                - int:
                    {'num_firms': (0, 100, 100000)}

                - dict or list, which should be strings of a dict or a
                  list (see example):
                    {'list_to_edit': "['brd', 'mlk', 'add']"}

                - a list of options:
                    {'several_options': ['opt_1', 'opt_2', 'opt_3']}

                - a string:
                    {'name': '2x2'}

        names (optional):
            a dictionary with the parameter name as key and an alternative
            text to be displayed instead.

        title:
            a string with the name of the simulation.

        header:
            html string for a bar on the top

        story:
            a dictionary with text to be displayed alongeside the graphs.
            Key must be the graphs title, value can be text or html.

        pages:
            A dictinoary with title as key and links to external websites
            as values, which are displayed on the right hand side.

        truncate_rounds:
            Does not display the initial x rounds, in the result graphs

        runtime:
            webbrowser to start the simulation in, can be 'xui' or python's
            webbrowser module's webrowser string.

        histograms:
            specifies in which round histograms are generated. If it is
            not specified rounds from the menu is used. Alternatively you can
            create 'histogram' parameter in parameter_mask.

        serve:
            If you run this on your local machine serve must be False.
            If used as a web server must be True

        hostname:
            Hostname if serve is active, defaults to '0.0.0.0'

        port:
            Port if serve is active, defaults to 80

        pypy:
            Name of the pypy interpreter to run ABCE super fast. e.G. 'pypy' or
            'pypy3'. The mainfile needs to be run with cpython e.G.:
            :code:`python3 start.py`

    Example::

        parameter_mask = {'name': 'name',
                          'random_seed': None,
                          'rounds': 40,
                          'num_firms': (0, 100, 100000),
                          'num_households': (0, 100, 100000),
                          'exponent': (0.0, 0.5, 1.1),
                          'several_options': ['opt_1', 'opt_2', 'opt_3']
                          'list_to_edit': "['brd', 'mlk', 'add']",
                          'dictionary_to_edit': "{'v1': 1, 'v2': 2}"}

        names = {'num_firms': 'Number of Firms'}

        @gui(parameter_mask, names,
             title="Agent-Based Computational Economics",
             serve=False)
        def main(simulation_parameters):
            parameter_list = eval(simulation_parameters['list_to_edit'])
            simulation = Simulation()
            firms = simulation.build_agents(Firm,
                simulation_parameters['num_firms'])
            households = simulation.build_agents(Household,
                simulation_parameters['num_households'])

            for r in range(simulation_parameters['rounds']):
                simulation.advance_round(r)
                firms.work()
                households.buy()

        if __name__ == '__main__':
            main(simulation_parameters)
    """
    parameter_mask = ({} if parameter_mask is None else parameter_mask)
    names = ({} if names is None else names)
    story = ({} if story is None else story)
    texts = ([abcedescription] if texts is None else texts)
    pages = ([] if pages is None else pages)

    if any('SPYDER' in name for name in os.environ):
        title = """WARNING: You are running this Simulation in SPYDER
                 When you run an IDE such as spyder the website
                 blocks. In order to avoid that, modify the 'Run Setting'
                 and choose 'Execute in external System Terminal' and
                 restart your system """
        print(title)

    def inner(simulation):
        if not IMPORTERROR:
            if pypy is not None:
                def simulation(parameters):
                    print("CALLING PYPY")
                    call([pypy,
                          sys.argv[0],
                          json.dumps(parameters),
                          abce.simulation_name])
            database = dataset.connect('sqlite:///parameter.db')
            abce.parameter_database = database['parameter']
            Form = form(parameter_mask, names)  # pylint: disable=C0103
            if serve:
                flexx.config.hostname = hostname
                flexx.config.port = port
                app.serve(basiclayout(Form, simulation, title, header,
                                      truncate_rounds,
                                      texts=texts, pages=pages,
                                      histograms=histograms))
                app.start()
            else:
                app.launch(basiclayout(Form, simulation, title, header,
                                       truncate_rounds,
                                       texts=texts, pages=pages,
                                       histograms=histograms),
                           windowmode='maximized', runtime=runtime)
                app.run()
        else:
            print("RUN PYPY")
            abce.simulation_name = sys.argv[2]
            simulation(json.loads(sys.argv[1]))
        return lambda _: None
    return inner


def graphs(parameter_mask=None, names=None):
    """ After the simulation simulation.graphs displays all logged data,
    this can not be use in conjuncture with @gui.

    Args:

        parameter_mask (optional):
            simulation parameters to display
        names (optional):
            a dictionary with the parameter name as key and an alternative
            text to be displayed instead.
    """
    names = ({} if names is None else names)

    if platform.python_implementation() == 'PyPy':
        try:
            call(['python3', '-mabce.show'])
        except Exception:
            call(['python', '-mabce.show'])
    else:

        database = dataset.connect('sqlite:///parameter.db')
        abce.parameter_database = database['parameter']
        parameter_mask = ({} if parameter_mask is None else parameter_mask)

        Form = form(parameter_mask, names)

        app.launch(basiclayout(Form, None, parameter_mask['name'],
                   graphs=True), runtime='browser-X')
        app.run()
